import logging
import re
import unicodedata
from enum import Enum


class NamingProfiles(str, Enum):
    # GraphQL naming profile ensures compatibility with the GraphQL specification
    # WRT names: https://spec.graphql.org/October2021/#Name
    graphql = "graphql"


class NameCompatibility:
    """Make a name compatible to the given profile"""

    # heading double underscores and digit reserved to names starting with a digit
    re_reserved_heading_digit = re.compile("^__[0-9]")
    # valid name between double underscores is reserved for unicode name transformations
    re_reserved_unicode_name_transformation = re.compile("__[0-9a-zA-Z][0-9a-zA-Z_]*__")
    # something like '__U_xxxx_' is reserved for unicode code transformations
    re_reserved_unicode_code_transformation = re.compile("__U_[0-9a-eA-E]{4}_")
    # strings starting with a digit
    re_heading_digit = re.compile("^[0-9]")
    # character that is neither alphanumeric nor underscore
    re_no_alphanum_underscore = re.compile("[^0-9a-zA-Z_]")

    def __init__(self, profile: NamingProfiles, do_not_fix: bool = False):
        """Specify the naming policy on instantiation"""
        self.profile = profile
        self.do_not_fix = do_not_fix

    def _no_heading_digits(self, input: str) -> str:
        """Ensure name does not start with a heading digit"""
        output = input
        if self.re_heading_digit.match(input):
            if self.do_not_fix:
                raise ValueError(f"Name '{input}' starts with digit (illegal GraphQL) and will not be fixed!")
            else:
                logging.warning(
                    f"Name '{input}' starts with digit (illegal GraphQL), "
                    + f"it has been prefixed with '__', resulting in {output}"
                )
            output = f"__{input}"
        return output

    def _transform_char(self, char: str) -> str:
        """Transform unsupported characters"""
        if len(char) != 1:
            raise Exception(f"Single character expected, but got '{char}'!!")
        # replace whitespaces with underscores
        # the transformation cannot be inverted, but is a well-established
        # and expected transformation
        if char == " ":
            return "_"
        # try to use names for ASCII characters
        if ord(char) < 128:
            try:
                # unicodedata.lookup should be able to invert the transformation
                return f"__{unicodedata.name(char).replace(' ', '_').replace('-', '_')}__"
            except ValueError:
                pass
        # fallback to code-transformation if none of the previous has worked
        return f"__U_{ord(char):04X}_"

    def _only_alphanum_underscore(self, input: str) -> str:
        """Ensure name does not contain any unsupported characters"""
        output = input
        # with re.split and re.findall we get in the same order and separated in two arrays
        # the substrings between special characters and the special characters
        no_alphanum_underscore_match = self.re_no_alphanum_underscore.findall(input)
        if no_alphanum_underscore_match:
            if self.do_not_fix:
                raise ValueError(f"Name '{input}' contains a character illegal in GraphQL and will not be fixed!")
            else:
                logging.warning(
                    f"Name '{input}' contains a character illegal in GraphQL, "
                    + f"the resulting encoded name is {output}"
                )
            to_keep = self.re_no_alphanum_underscore.split(input)
            # first comes first substring to keep
            output = to_keep[0]
            # each char replacement is followed by the next substring to keep
            for offset in range(0, len(to_keep) - 1):
                output = output + self._transform_char(no_alphanum_underscore_match[offset])
                output = output + to_keep[offset + 1]
        return output

    def _graphql_compatibility(self, input: str) -> str:
        """Ensure name compatibility with GraphQL name policies"""
        # as of now, some (hopefully) very rare patterns are reserved to mark transformations
        if self.re_reserved_heading_digit.match(input):
            raise NotImplementedError("Names starting with a double underscore followed by a digit are not supported!!")
        if self.re_reserved_unicode_name_transformation.match(input):
            raise NotImplementedError("Names containing valid names between double underscores are not supported!!")
        if self.re_reserved_unicode_code_transformation.match(input):
            raise NotImplementedError("Names containing strings like '__U_xxxx_' are not supported!!")
        # apply transformation
        output = input
        output = self._no_heading_digits(output)
        output = self._only_alphanum_underscore(output)
        return output

    def compatible(self, input: str) -> str:
        """Make given name compatible with the given naming policy."""
        if self.profile == "graphql":
            return self._graphql_compatibility(input)



"""Validate linkml input and optionally emit completely resolved biolink yaml output"""

import os
from dataclasses import dataclass

import click
from linkml_runtime.utils.yamlutils import as_yaml

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments
from linkml.utils.schemaloader import load_raw_schema


@dataclass
class YAMLGenerator(Generator):
    """
    A generator that produces a schema as a YAML Document
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.0"
    valid_formats = ["yaml"]
    uses_schemaloader = True

    # ObjectVars
    validateonly: bool = False

    def serialize(self, validateonly: bool = False, **kwargs) -> str:
        if validateonly:
            return self.synopsis.summary()
        else:
            return as_yaml(self.schema)


@shared_arguments(YAMLGenerator)
@click.command(name="yaml")
@click.option(
    "--raw/--no-raw",
    default=False,
    show_default=True,
    help="Use the raw loader and do not inject additional information",
)
@click.option(
    "--validateonly/--generate",
    "-V/-g",
    default=False,
    show_default=True,
    help="Just validate / generate output (default: generate)",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, raw: bool, **args):
    """Validate input and produce fully resolved yaml equivalent"""
    if raw:
        s = load_raw_schema(yamlfile)
        print(as_yaml(s))
    else:
        gen = YAMLGenerator(yamlfile, **args)
        print(gen.serialize(**args))


if __name__ == "__main__":
    cli()



"""
The ``linkml.validator.loaders`` package contain classes which are responsible for yielding data
instance from a source. Instances of these classes are passed to
:meth:`linkml.validator.Validator.validate_source` and
:meth:`linkml.validator.Validator.iter_results_from_source`
"""

import os
from typing import Union

from linkml.validator.loaders.delimited_file_loader import CsvLoader, TsvLoader
from linkml.validator.loaders.json_loader import JsonLoader
from linkml.validator.loaders.loader import Loader
from linkml.validator.loaders.yaml_loader import YamlLoader


def default_loader_for_file(file: Union[str, bytes, os.PathLike]) -> Loader:
    _, ext = os.path.splitext(file)
    if ext == ".csv":
        return CsvLoader(file, skip_empty_rows=True)
    elif ext == ".tsv":
        return TsvLoader(file, skip_empty_rows=True)
    elif ext == ".json":
        return JsonLoader(str(file))
    elif ext in (".yaml", ".yml"):
        return YamlLoader(file)

    raise ValueError(f"Could not find loader for file: {file}")


__all__ = [
    "CsvLoader",
    "JsonLoader",
    "Loader",
    "TsvLoader",
    "YamlLoader",
    "default_loader_for_file",
]



"""
A model transformer that transforms a schema into logical form.

Formally, this transformer generates a logical model of the schema in which inheritance
of slots/attributes is replaced by boolean conjunctions, and the final form is simplified.

For example, given a slot `s`:

```
slots:
  s:
    range: integer
    minimum_value: 1
```

Which is reused and refined by C and C (where C is_a D):

```
classes:
  D:
    slots:
     - s
    slot_usage:
      s:
         minimum_value: 2
  C:
    is_a: D
    slot_usage:
      s:
        minimum_value: 3
```

The hierarchy unrolling step will create an attribute `s` for C that is the conjunction of usages for that
class, ancestors, and the base slot:

```
s:
    range: integer
    all_of:
        - minimum_value: 1
        - minimum_value: 2
        - minimum_value: 3
```

This is then simplified using symbolic reasoning to:

```
s:
    range: integer
    minimum_value: 3
```

See logictools.py for the symbolic reasoning engine.



"""

import logging
from collections.abc import Iterator
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Callable, Optional, Union

from linkml_runtime import SchemaView
from linkml_runtime.dumpers import json_dumper
from linkml_runtime.linkml_model import ClassDefinitionName, SchemaDefinition
from linkml_runtime.linkml_model.meta import (
    AnonymousSlotExpression,
    AnonymousTypeExpression,
    SlotDefinition,
    SlotDefinitionName,
    TypeDefinition,
)

from linkml.transformers.model_transformer import ModelTransformer
from linkml.utils import logictools

HERITABLE_METASLOT = [
    # "multivalued",
    "range",
    "range_expression",
    "minimum_value",
    "maximum_value",
    "minimum_cardinality",
    "maximum_cardinality",
    "pattern",
    "structured_pattern",
    "required",
    "recommended",
    "inlined",
    "inlined_as_list",
    "all_of",
    "exactly_one_of",
    "none_of",
    "any_of",
    "unit",
    "value_presence",
    "equals_string",
    "equals_string_in",
    "equals_number",
    "equals_expression",
    "has_member",
    "all_members",
]
"""Metamodel slots that can be rolled down or inherited"""

DIRECT_ROLL_DOWN = ["multivalued", "key", "identifier", "designates_type"]
"""Additional slots that can be rolled down, but only if they are on a SlotDefinition, not a slot expression"""

HERITABLE_TYPE_METASLOT = [
    "minimum_value",
    "maximum_value",
    # "minimum_cardinality",
    # "maximum_cardinality",
    "pattern",
    "structured_pattern",
    "all_of",
    "exactly_one_of",
    "none_of",
    "any_of",
    "unit",
    "repr",
]

MATCHES_PREDICATE = "matches"

logger = logging.getLogger(__name__)


class UnsatisfiableAttribute(ValueError):
    """The attribute specifies constraints that cannot be satisfied"""

    pass


@dataclass
class LogicalModelTransformer(ModelTransformer):
    """
    Flatten a schema by rolling down slots and attributes to subclasses.

    The primary application of this is as an intermediate product to be used in generators.
    For example, when compiling LinkML to frameworks that either lack inheritance (e.g. JSON-Schema),
    or only partially support inheritance (e.g. single inheritance), this transformer can be used
    to generate a schema with attributes that are as complete as possible within that framework.

    To demonstrate, consider a simple schema with a class Person inheriting from Thing:

    >>> from linkml.utils.schema_builder import SchemaBuilder
    >>> sb = SchemaBuilder()
    >>> _ = sb.add_class("Thing", slots=["id", "name"])
    >>> _ = sb.add_class("Person", slots={"age": {"range": "integer"}}, is_a="Thing")
    >>> _ = sb.add_class("Organization", slots={"category": {"range": "OrganizationType"}}, is_a="Thing")
    >>> _ = sb.add_enum("OrganizationType", ["commercial", "non-profit"])
    >>> _ = sb.add_defaults()
    >>> from linkml.transformers.logical_model_transformer import LogicalModelTransformer
    >>> tr = LogicalModelTransformer()
    >>> tr.set_schema(sb.schema)
    >>> flat_schema = tr.transform()
    >>> from linkml_runtime.dumpers import yaml_dumper
    >>> print(yaml_dumper.dumps(flat_schema.classes['Person']))
    name: Person
    attributes:
      id:
        name: id
        range: string
      name:
        name: name
        range: string
      age:
        name: age
        range: integer

    ...

    Note that the id and name attributes have been rolled down from Thing to Person.

    This transformer will translate range assertions into any_of assertions, where the class
    hierarchy is rolled up:

    >>> _ = sb.add_class("Container", slots={"entities": {"range": "Thing", "multivalued": True}})
    >>> tr.set_schema(sb.schema)
    >>> flat_schema = tr.transform()
    >>> print(yaml_dumper.dumps(flat_schema.classes['Container']))
    name: Container
    attributes:
      entities:
        name: entities
        multivalued: true
        any_of:
        - range: Organization
        - range: Person
        - range: Thing

    ...

    This has "compiled away" the hierarchy. The type of the values of 'entities' for Container can
    be checked against the any_of expression, without needing to know the full class hierarchy.
    This is useful when compiling to a framework that does not support inheritance.

    If you are compiling to a framework that does have inheritance, then you can specify that these are
    preserved:

    >>> tr_inh = LogicalModelTransformer(preserve_class_mixins=True, preserve_class_is_a=True)
    >>> tr_inh.set_schema(sb.schema)
    >>> flat_schema = tr_inh.transform()
    >>> print(yaml_dumper.dumps(flat_schema.classes['Container']))
    name: Container
    from_schema: http://example.org/test-schema
    attributes:
      entities:
        name: entities
        range: Thing
        multivalued: true

    ...

    You can also use any_of directly in the schema:

    >>> _ = sb.add_class("Container2", slots={"entities2": {"multivalued": True,
    ...                                                     "any_of": [{"range": "Person"},
    ...                                                                {"range": "Organization"}]}})
    >>> tr.set_schema(sb.schema)
    >>> flat_schema = tr.transform()
    >>> print(yaml_dumper.dumps(flat_schema.classes['Container2']))
    name: Container2
    attributes:
      entities2:
        name: entities2
        multivalued: true
        any_of:
        - range: Organization
        - range: Person

    ...

    Note that defaults are always applied after reasoning. So if you set a default_range to
    be string, and use inheritance preservation, the unfolded range expression is still
    correct:

    >>> sb.schema.default_range = "string"
    >>> tr_inh.set_schema(sb.schema)
    >>> flat_schema = tr_inh.transform()
    >>> print(yaml_dumper.dumps(flat_schema.classes['Container2']))
    name: Container2
    from_schema: http://example.org/test-schema
    attributes:
      entities2:
        name: entities2
        multivalued: true
        any_of:
        - range: Organization
        - range: Person


    When compiling to a framework that does not support inheritance, such as JSON Schema, the any_of expression
    can be directly translated to the same construct in JSON-Schema.

    This transformer will also perform reasoning on the schema, to provide the simplest complete form. Let's add
    some constraints onto the "age" slot and create a subclass of Person with a constrained age range:

    >>> sb.schema.slots["age"].minimum_value = 0
    >>> sb.schema.slots["age"].maximum_value = 200
    >>> _ = sb.add_class("MiddleAged", slot_usage={"age": {"minimum_value": 40, "maximum_value": 60}}, is_a="Person")
    >>> tr.set_schema(sb.schema)
    >>> flat_schema = tr.transform()
    >>> print(yaml_dumper.dumps(flat_schema.classes['MiddleAged'].attributes['age']))
    name: age
    range: integer
    minimum_value: 40
    maximum_value: 60

    Here the slot-level constraints seem to have been "overridden". In fact, LinkML is monotonic, and there
    is no overriding. In fact the constraints have been combined as a conjunction (And), and then the resulting
    entailed attribute has been *simplified* (age >= 40 & age >= 0 ==> age >= 40).

    The reasoning engine can also detect unsatisfiable constraints:

    >>> from linkml.transformers.logical_model_transformer import UnsatisfiableAttribute
    >>> _ = sb.add_class("YoungAdult", slot_usage={"age": {"minimum_value": 40, "maximum_value": 10}}, is_a="Person")
    >>> tr.set_schema(sb.schema)
    >>> try:
    ...   flat_schema = tr.transform()
    ... except UnsatisfiableAttribute as e:
    ...   print(e)
    <BLANKLINE>
    Attribute YoungAdult.age is unsatisfiable
    <BLANKLINE>

    Let's get rid of the problematic class:

    >>> del sb.schema.classes["YoungAdult"]

    And try a different example - we will try and override the range of the age slot with a string:

    >>> _ = sb.add_class("Person2", is_a="Person", slot_usage={"age": {"range": "string"}})
    >>> tr.set_schema(sb.schema)
    >>> try:
    ...   flat_schema = tr.transform()
    ... except UnsatisfiableAttribute as e:
    ...   print(e)
    <BLANKLINE>
    Attribute Person2.age is unsatisfiable
    <BLANKLINE>

    """

    preserve_class_is_a: bool = False
    """
    If True, preserve the is_a hierarchy in the flattened schema, and assume that entailments
    will hold in the transformed schema. If False, the is_a hierarchy will be flattened.
    """

    preserve_class_mixins: bool = False
    """
    If True, preserve the mixin hierarchy in the flattened schema, and assume that entailments
    will hold in the transformed schema. If False, the mixin hierarchy will be flattened.
    """

    preserve_type_is_a: bool = False
    """
    If True, preserve the is_a hierarchy in the flattened schema, and assume that entailments
    will hold in the transformed schema. If False, the is_a hierarchy will be flattened.
    """

    tidy_slots: bool = True
    """
    If True, remove all standalone slots from transformed schema. These will have been rolled
    up into class attributes.
    """

    tidy_inheritance: bool = True
    """
    If True, remove all inheritance from transformed schema. These will have been rolled
    up into class attributes.
    """

    tidy_default_range: bool = True
    """
    If True, remove the fake "Any" class from the transformed schema. This will be implicit based on
    a lack of range assignment.
    """

    reason_over_metamodel_slots: Optional[list[str]] = None
    """
    If set, only reason over the specified metamodel slots.
    """

    def transform(self, tgt_schema_name: str = None, simplify=True, force_any_of=False, **kwargs) -> Any:
        """
        Transform the schema using logical reasoning, materializing entailed simple attributes.

        :param tgt_schema_name:
        :param simplify: If True (default), simplify the schema after transformation
        :param force_any_of: If True, force the use of any_of expressions for range, even for singletons
        :return:
        """
        sv = self.schemaview
        target_schemaview = SchemaView(deepcopy(self.source_schema))
        target_schemaview.merge_imports()
        target_schema = target_schemaview.schema
        for tn, typ in target_schema.types.items():
            ancs = sv.type_ancestors(tn, reflexive=False)
            logger.debug(f"Unrolling type {tn}, merging {len(ancs)}")
            if ancs:
                for type_anc in ancs:
                    self._merge_type_ancestors(target=typ, source=sv.get_type(type_anc))
        for sn, slot in target_schema.slots.items():
            ancs = sv.slot_ancestors(sn, reflexive=False)
            logger.debug(f"Unrolling slot {sn}, merging {len(ancs)}")
            if ancs:
                for slot_anc in ancs:
                    self._merge_slot_ancestors(target=slot, source=target_schema.slots[slot_anc])
        for cn, cls in target_schema.classes.items():
            ancs = sv.class_ancestors(
                cn,
                reflexive=True,
                is_a=not self.preserve_class_is_a,
                mixins=not self.preserve_class_mixins,
                depth_first=False,
            )
            ancs = list(reversed(ancs))
            logger.debug(f"Unrolling class {cn}, merging {len(ancs)}")
            self._roll_down(target_schema, cn, ancs)
        self.apply_defaults(target_schema)
        if simplify:
            self.simplify(target_schema, force_any_of=force_any_of)
        if self.tidy_slots:
            target_schema.slots = {}
        if self.tidy_inheritance:
            for c in target_schema.classes.values():
                if not self.preserve_class_is_a:
                    c.is_a = None
                if not self.preserve_class_mixins:
                    c.mixins = []
            if not self.preserve_type_is_a:
                for t in target_schema.types.values():
                    t.is_a = None
        if self.tidy_default_range:
            target_schema.default_range = None
            any_proxy_classes = [c for c in target_schema.classes.values() if c.class_uri == "linkml:Any"]
            if any_proxy_classes:
                for c in any_proxy_classes:
                    del target_schema.classes[c.name]
        return target_schema

    def _roll_down(
        self,
        target_schema: SchemaDefinition,
        target_class_name: ClassDefinitionName,
        ancestors: list[ClassDefinitionName],
    ):
        anc_classes = [self.schemaview.get_class(ancestor) for ancestor in ancestors]
        attributes: dict[SlotDefinitionName, SlotDefinition] = {}
        for ancestor_class in anc_classes:
            top_level_slots = [(s, target_schema.slots[s]) for s in ancestor_class.slots]
            for slot_name, slot_expr in (
                list(ancestor_class.attributes.items()) + list(ancestor_class.slot_usage.items()) + top_level_slots
            ):
                if slot_name not in attributes:
                    attributes[slot_name] = SlotDefinition(slot_name)
                sx = attributes[slot_name]
                self._merge_slot_ancestors(target=sx, source=slot_expr)
        tgt_cls = target_schema.classes[target_class_name]
        for k, v in attributes.items():
            tgt_cls.attributes[k] = v
        tgt_cls.slots = []
        tgt_cls.slot_usage = {}

    def _merge_slot_ancestors(
        self,
        target: Union[AnonymousSlotExpression, SlotDefinition],
        source: Union[SlotDefinition, AnonymousSlotExpression],
    ):
        """
        Generate all_ofs for slot based on ancestors.

        :param target:
        :param source:
        :return:
        """
        new_slot_dict = {}
        heritable_metaslots = self.reason_over_metamodel_slots or HERITABLE_METASLOT
        for p in heritable_metaslots:
            pv = getattr(source, p)
            if pv is not None and pv != [] and pv != {} and pv is not False:
                new_slot_dict[p] = pv
        for p in DIRECT_ROLL_DOWN:
            pv = getattr(source, p, None)
            if pv is not None and pv != [] and pv != {} and pv is not False:
                setattr(target, p, pv)
        sv = self.schemaview
        if source.range in sv.all_types():
            typ = sv.get_type(source.range)
            for p in set(HERITABLE_TYPE_METASLOT).intersection(heritable_metaslots):
                pv = getattr(typ, p)
                if pv is not None and pv != [] and pv != {} and pv is not False:
                    new_slot_dict[p] = pv
        if new_slot_dict:
            target.all_of.append(AnonymousSlotExpression(**new_slot_dict))

    @staticmethod
    def _merge_type_ancestors(target: TypeDefinition, source: TypeDefinition):
        """
        Generate all_ofs for type based on ancestors.

        :param target:
        :param source:
        :return:
        """
        new_slot_dict = {}
        for p in HERITABLE_TYPE_METASLOT:
            if p in ["repr"]:
                # not on anonymous types
                continue
            pv = getattr(source, p)
            if pv is not None and pv != [] and pv != {} and pv is not False:
                new_slot_dict[p] = pv
        if new_slot_dict:
            target.all_of.append(AnonymousTypeExpression(**new_slot_dict))

    def apply_defaults(self, target_schema: SchemaDefinition):
        """
        Apply defaults to the schema

        :param target_schema:
        :return:
        """
        for cls in target_schema.classes.values():
            for att in cls.attributes.values():
                self._apply_defaults_to_attribute(att)

    def _apply_defaults_to_attribute(self, att: SlotDefinition):
        if not att.range:
            if not any(self._collection_slot_expressions(att, lambda sx: sx.range)):
                att.range = self.schemaview.schema.default_range

    def _collection_slot_expressions(
        self, att: Union[SlotDefinition, AnonymousSlotExpression], filter_function: Callable
    ) -> Iterator[AnonymousSlotExpression]:
        """
        Traverse nested slot expressions yielding those that match the filter function.

        :param att:
        :return:
        """
        if filter_function(att):
            yield att
        for sx in att.all_of + att.exactly_one_of + att.none_of + att.any_of:
            yield from self._collection_slot_expressions(sx, filter_function)

    def simplify(self, target_schema: SchemaDefinition, force_any_of=False):
        for cls in target_schema.classes.values():
            for att in cls.attributes.values():
                x = self._as_logical_expression(att)
                logger.debug(f"Translating {cls.name}.{att.name} ==> {x}")
                x = logictools.simplify_full(x)
                logger.debug(f"Simplified: {x}")
                if logictools.is_contradiction(x):
                    raise UnsatisfiableAttribute(f"Attribute {cls.name}.{att.name} is unsatisfiable")
                self._simplify_member_ofs(x)
                if force_any_of:
                    if not isinstance(x, logictools.Or):
                        x = logictools.Or(x)
                logger.debug(f"Simplified member of: {x}")
                simplified_att = self._from_logical_expression(x)
                for k, v in simplified_att.__dict__.items():
                    if v is None or v is False:
                        if k in ["multivalued"]:
                            continue
                    setattr(att, k, v)

    def _simplify_member_ofs(self, expr: logictools.Expression):
        if isinstance(expr, logictools.Term) and expr.predicate == "in":
            elements = expr.operands[1]
            if elements == logictools.UniversalSet():
                pass
            else:
                expr.operands[1] = self._remove_redundant(elements)
        elif isinstance(expr, logictools.And):
            [self._simplify_member_ofs(o) for o in expr.operands]
        elif isinstance(expr, logictools.Or):
            [self._simplify_member_ofs(o) for o in expr.operands]
        elif isinstance(expr, logictools.Not):
            self._simplify_member_ofs(expr.operand)

    def _remove_redundant(self, elements: list[str]) -> list[str]:
        sv = self.schemaview
        redundant = set()
        if not self.preserve_class_mixins:
            redundant.update([c.name for c in sv.all_classes().values() if c.mixin])
            if not self.preserve_class_is_a:
                redundant.update([c.name for c in sv.all_classes().values() if c.abstract])
        for x in elements:
            if x in sv.all_classes():
                redundant.update(
                    sv.class_descendants(
                        x,
                        reflexive=False,
                        is_a=self.preserve_class_is_a,
                        mixins=self.preserve_class_mixins,
                    )
                )
            elif x in sv.all_types():
                redundant.update(self._type_descendants(x, reflexive=False))
            elif x in sv.all_enums():
                redundant.update(self._enum_descendants(x, reflexive=False))
            else:
                logger.warning(f"Unknown class {x} in {elements}")
        return [x for x in elements if x not in redundant]

    def _type_descendants(self, type_name: str, imports=True, reflexive=True, depth_first=True) -> list[str]:
        # TODO: move this to schemaview
        sv = self.schemaview
        from linkml_runtime.utils.schemaview import _closure

        def type_children(type_name, imports=True):
            elts = [sv.get_type(x) for x in sv.all_types(imports)]
            return [x.name for x in elts if x.typeof == type_name]

        return _closure(
            lambda x: type_children(x, imports=imports),
            type_name,
            reflexive=reflexive,
            depth_first=depth_first,
        )

    def _enum_descendants(self, enum_name: str, imports=True, reflexive=True, depth_first=True) -> list[str]:
        # TODO: move this to schemaview
        sv = self.schemaview
        from linkml_runtime.utils.schemaview import _closure

        def enum_children(enum_name, imports=True):
            elts = [sv.get_enum(x) for x in sv.all_enums(imports)]
            return [x.name for x in elts if x.is_a == enum_name]

        return _closure(
            lambda x: enum_children(x, imports=imports),
            enum_name,
            reflexive=reflexive,
            depth_first=depth_first,
        )

    @property
    def _value_var(self) -> logictools.Variable:
        return logictools.Variable("value")

    @property
    def _length_var(self) -> logictools.Variable:
        return logictools.Variable("length")

    @property
    def _type_var(self) -> logictools.Variable:
        return logictools.Variable("type")

    def _as_logical_expression(
        self, slot_expression: Union[SlotDefinition, AnonymousSlotExpression]
    ) -> logictools.Expression:
        sv = self.schemaview
        exprs = []
        value_var = self._value_var
        type_var = self._type_var
        length_var = self._length_var
        if slot_expression.minimum_value is not None:
            exprs.append(value_var >= slot_expression.minimum_value)
        if slot_expression.maximum_value is not None:
            exprs.append(value_var <= slot_expression.maximum_value)
        if slot_expression.minimum_cardinality is not None:
            exprs.append(length_var >= slot_expression.minimum_cardinality)
        if slot_expression.maximum_cardinality is not None:
            exprs.append(length_var <= slot_expression.maximum_cardinality)
        if slot_expression.pattern is not None:
            exprs.append(logictools.Term(MATCHES_PREDICATE, value_var, slot_expression.pattern))
        if slot_expression.range:
            rn = slot_expression.range
            if slot_expression.range in sv.all_classes():
                rn_cls = sv.get_class(rn)
                if rn_cls.class_uri == "linkml:Any":
                    elts = logictools.UniversalSet()
                else:
                    elts = sv.class_descendants(rn, reflexive=True)
            elif slot_expression.range in sv.all_types():
                elts = self._type_descendants(rn, reflexive=True)
            else:
                elts = [rn]
            exprs.append(logictools.IsIn(type_var, elts))
        if slot_expression.all_of:
            exprs.append(logictools.And(*[self._as_logical_expression(subx) for subx in slot_expression.all_of]))
        if slot_expression.exactly_one_of:
            # TODO: disjointedness
            exprs.append(logictools.Or(*[self._as_logical_expression(subx) for subx in slot_expression.exactly_one_of]))
        if slot_expression.none_of:
            exprs.append(
                logictools.Not(logictools.Or(*[self._as_logical_expression(subx) for subx in slot_expression.none_of]))
            )
        if slot_expression.any_of:
            exprs.append(logictools.Or(*[self._as_logical_expression(subx) for subx in slot_expression.any_of]))
        for p in HERITABLE_METASLOT:
            if p in [
                "range",
                "pattern",
                "minimum_value",
                "maximum_value",
                "all_of",
                "exactly_one_of",
                "none_of",
                "any_of",
            ]:
                # already accounted for
                continue
            v = getattr(slot_expression, p, None)
            if v is not None and v != [] and v != {}:
                exprs.append(logictools.Eq(p, v))
        return logictools.And(*exprs)

    def _from_logical_expression(self, expr: logictools.Expression) -> AnonymousSlotExpression:
        if isinstance(expr, logictools.And):
            all_of = [self._from_logical_expression(e) for e in expr.operands]
            if len(all_of) == 1:
                return all_of[0]
            x = json_dumper.to_dict(all_of[0])
            remaining_all_of = []
            for y in all_of[1:]:
                y = json_dumper.to_dict(y)
                if set(x.keys()).intersection(set(y.keys())):
                    remaining_all_of.append(AnonymousSlotExpression(**y))
                else:
                    x.update(y)
            ase = AnonymousSlotExpression(**x)
            ase.all_of.extend(remaining_all_of)
            return ase
        elif isinstance(expr, logictools.Or):
            return AnonymousSlotExpression(any_of=[self._from_logical_expression(e) for e in expr.operands])
        elif isinstance(expr, logictools.Not):
            return AnonymousSlotExpression(none_of=[self._from_logical_expression(expr.operand)])
        elif isinstance(expr, logictools.Eq):
            var = expr.lhs
            val = expr.rhs
            return AnonymousSlotExpression(**{var: val})
        # elif isinstance(expr, logictools.IsIn):
        #    return AnonymousSlotExpression(range=expr.operands[1])
        elif isinstance(expr, logictools.Term):
            var = expr.operands[0]
            val = expr.operands[1]
            if var == self._value_var:
                if expr.predicate == MATCHES_PREDICATE:
                    return AnonymousSlotExpression(pattern=val)
                elif expr.predicate == ">=":
                    return AnonymousSlotExpression(minimum_value=val)
                elif expr.predicate == ">":
                    return AnonymousSlotExpression(minimum_value=val + 1)
                elif expr.predicate == "<=":
                    return AnonymousSlotExpression(maximum_value=val)
                elif expr.predicate == "<":
                    return AnonymousSlotExpression(maximum_value=val - 1)
                else:
                    raise ValueError(f"Unknown predicate {expr.predicate} for {var}")
            elif var == self._length_var:
                if expr.predicate == ">=":
                    return AnonymousSlotExpression(minimum_cardinality=val)
                elif expr.predicate == ">":
                    return AnonymousSlotExpression(minimum_cardinality=val + 1)
                elif expr.predicate == "<=":
                    return AnonymousSlotExpression(maximum_cardinality=val)
                elif expr.predicate == "<":
                    return AnonymousSlotExpression(maximum_cardinality=val - 1)
                else:
                    raise ValueError(f"Unknown predicate {expr.predicate} for {var}")
            elif var == self._type_var:
                if expr.predicate == "in":
                    if val == logictools.UniversalSet():
                        return AnonymousSlotExpression(range="Any")
                    elif len(val) == 1:
                        return AnonymousSlotExpression(range=val[0])
                    else:
                        return AnonymousSlotExpression(any_of=[AnonymousSlotExpression(range=v) for v in sorted(val)])
                else:
                    raise ValueError(f"Unknown predicate {expr.predicate} for {var}")
            else:
                raise ValueError(f"Unknown variable {var}")
        return AnonymousSlotExpression()

    def attribute_as_python_field(self, attribute: SlotDefinition) -> str:
        """
        Return a Python field declaration for the given attribute.

        Note: this method may move to pydanticgen, it is currently here for demonstration purposes.

        :param attribute:
        :return:
        """
        return f"{attribute.name}: {self._att_as_python_expression(attribute)}"

    def _att_as_python_expression(
        self,
        attribute: Union[SlotDefinition, AnonymousSlotExpression],
        root_slot: SlotDefinition = None,
        stack: list = None,
    ) -> str:
        # Note: in future versions of the metamodel, multivalued may move to the expression
        if stack is None:
            stack = [attribute]
        if root_slot is None:
            if isinstance(attribute, SlotDefinition):
                root_slot = attribute
            else:
                raise ValueError(f"root_slot must be provided for {attribute}")
        multivalued = root_slot.multivalued
        s = ""
        if attribute.any_of:
            union_terms = [self._att_as_python_expression(x, root_slot, stack + [x]) for x in attribute.any_of]
            s = f"Union[{', '.join(union_terms)}]"
        else:
            if attribute.range:
                s = attribute.range
            else:
                s = "Any"
            if not attribute.required:
                s = f"Optional[{s}]"
            if multivalued:
                if attribute.inlined and not attribute.inlined_as_list:
                    s = f'Dict["str", {s}]'
                else:
                    # TODO: ref by key
                    s = f"Collection[{s}]"
        return s



import logging
import os
import re
from dataclasses import dataclass

import click
from linkml_runtime.linkml_model.meta import ClassDefinition, EnumDefinition, SlotDefinition
from linkml_runtime.utils.formatutils import camelcase, lcamelcase

from linkml._version import __version__
from linkml.generators.common.naming import NameCompatibility, NamingProfiles
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class GraphqlGenerator(Generator):
    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["graphql"]
    visit_all_class_slots = True
    uses_schemaloader = True
    requires_metamodel = False

    strict_naming: bool = False
    _permissible_value_valid_characters = re.compile("^[_A-Za-z][_0-9A-Za-z]*?$")
    _types_any = []

    def __post_init__(self):
        self.name_compatiblity = NameCompatibility(profile=NamingProfiles.graphql, do_not_fix=self.strict_naming)
        super().__post_init__()

    def visit_schema(self, **kwargs) -> str:
        return self.generate_header()

    def generate_header(self) -> str:
        out = f"# metamodel_version: {self.schema.metamodel_version}\n"
        if self.schema.version:
            out += f"# version: {self.schema.version}\n"
        return out

    def visit_class(self, cls: ClassDefinition) -> str:
        # no type can be declared for subtypes of "Any"
        if cls.class_uri == "linkml:Any":
            self._types_any.append(cls.name)
            return f"scalar {cls.name}"
        etype = "interface" if (cls.abstract or cls.mixin) and not cls.mixins else "type"
        mixins = ", ".join([camelcase(mixin) for mixin in cls.mixins])
        out = f"{etype} {camelcase(cls.name)}" + (f" implements {mixins}" if mixins else "")
        out = "\n".join([out, "  {"])
        return out

    def end_class(self, cls: ClassDefinition) -> str:
        if cls.name in self._types_any:
            return "\n\n"
        else:
            return "\n  }\n\n"

    def visit_class_slot(self, cls: ClassDefinition, aliased_slot_name: str, slot: SlotDefinition) -> str:
        if slot.range in self.schema.classes or slot.range in self.schema.slots or slot.range in self.schema.enums:
            slotrange = camelcase(slot.range)
        elif slot.range in self.schema.types:
            if self.schema.types[slot.range].from_schema != "https://w3id.org/linkml/types":
                slotrange = camelcase(slot.range)
            else:
                graphql_scalars = ["Int", "Float", "String", "Boolean", "ID"]
                if slot.range == "integer":
                    slotrange = "Int"
                elif slot.range == "decimal":
                    slotrange = "Float"
                elif camelcase(slot.range) in graphql_scalars:
                    slotrange = camelcase(slot.range)
                else:
                    if self.schema.types[slot.range].repr:
                        python_type = self.schema.types[slot.range].repr
                    elif self.schema.types[slot.range].base:
                        python_type = self.schema.types[slot.range].base
                    if str(python_type) == "float":
                        slotrange = "Float"
                    elif str(python_type) == "str":
                        slotrange = "String"

        if slot.multivalued:
            slotrange = f"[{slotrange}]"
        if slot.required:
            slotrange = slotrange + "!"
        return f"\n    {lcamelcase(aliased_slot_name)}: {slotrange}"

    def visit_enum(self, enum: EnumDefinition):
        if enum.permissible_values:
            permissible_values = []
            for value in enum.permissible_values:
                permissible_values.append(self.name_compatiblity.compatible(value))
            values = "\n    ".join(permissible_values)
            return f"enum {camelcase(enum.name).replace(' ', '')}\n  {{\n    {values}\n  }}\n\n"
        else:
            logging.warning(
                f"Enumeration {enum.name} using `reachable_from` instead of `permissible_values` "
                + "to specify permissible values is not supported yet."
                + "Enumeration {enum.name} will be silently ignored!!"
            )
            return ""


@shared_arguments(GraphqlGenerator)
@click.command(name="graphql")
@click.option(
    "--strict-naming",
    is_flag=True,
    show_default=True,
    help="Treat warnings about invalid names or schema elements as errors.",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, **args):
    """Generate graphql representation of a LinkML model"""
    generator = GraphqlGenerator(yamlfile, **args)
    print(generator.serialize(**args))


if __name__ == "__main__":
    cli()



import os
from collections.abc import Iterator
from typing import Any, Optional

from jsonschema.exceptions import best_match

from linkml.validator.plugins.validation_plugin import ValidationPlugin
from linkml.validator.report import Severity, ValidationResult
from linkml.validator.validation_context import ValidationContext


class JsonschemaValidationPlugin(ValidationPlugin):
    """A validation plugin which validates instances using a JSON Schema validator.

    :param closed: If ``True``, additional properties are not allowed on instances.
        Defaults to ``False``.
    :param include_range_class_descendants: If True, use an open world assumption and allow the
        range of a slot to be any descendant of the declared range. Note that if the range of a
        slot has a type designator, descendants will always be included.
    :param json_schema_path: If provided, JSON Schema will not be generated from the schema,
        instead it will be read from this path. In this case the value of the ``closed`` argument
        is disregarded and the open- or closed-ness of the existing JSON Schema is taken as-is.
    """

    def __init__(
        self,
        *,
        closed: bool = False,
        include_range_class_descendants: bool = True,
        json_schema_path: Optional[os.PathLike] = None,
    ) -> None:
        self.closed = closed
        self.include_range_class_descendants = include_range_class_descendants
        self.json_schema_path = json_schema_path

    def process(self, instance: Any, context: ValidationContext) -> Iterator[ValidationResult]:
        """Perform JSON Schema validation on the provided instance

        :param instance: The instance to validate
        :param context: The validation context which provides a JSON Schema artifact
        :return: Iterator over validation results
        :rtype: Iterator[ValidationResult]
        """
        validator = context.json_schema_validator(
            closed=self.closed,
            include_range_class_descendants=self.include_range_class_descendants,
            path_override=self.json_schema_path,
        )
        for error in validator.iter_errors(instance):
            error_context = [ctx.message for ctx in error.context]
            best_error = best_match([error])
            yield ValidationResult(
                type="jsonschema validation",
                severity=Severity.ERROR,
                instance=instance,
                instantiates=context.target_class,
                message=f"{best_error.message} in /{'/'.join(str(p) for p in best_error.absolute_path)}",
                context=error_context,
                source=best_error,
            )



from collections import defaultdict
from typing import IO, Any, Optional

from linkml.linter.config.datamodel.config import RuleLevel
from linkml.linter.linter import LinterProblem

from .formatter import Formatter


class MarkdownFormatter(Formatter):
    def __init__(self, file: Optional[IO[Any]] = None) -> None:
        super().__init__(file)
        self.total = 0
        self.total_errors = 0
        self.total_warnings = 0
        self.schemas_with_error = set()
        self.schemas_with_warning = set()
        self.problems = defaultdict(list)

    def start_schema(self, name: str):
        self.total += 1
        self.current_schema = name

    def handle_problem(self, problem: LinterProblem):
        if str(problem.level) == RuleLevel.warning.text:
            self.total_warnings += 1
            self.schemas_with_warning.add(self.current_schema)
        elif str(problem.level) == RuleLevel.error.text:
            self.total_errors += 1
            self.schemas_with_error.add(self.current_schema)
        self.problems[self.current_schema].append(problem)

    def write_summary(self):
        self.write(
            f"""
## Summary

|                      | Count |
|----------------------|-------|
| Schemas Checked      | {self.total} |
| Schemas with Error   | {len(self.schemas_with_error)} |
| Schemas with Warning | {len(self.schemas_with_warning)} |
| Total Errors         | {self.total_errors} |
| Total Warnings       | {self.total_warnings} |

"""
        )

    def write_details(self):
        if len(self.problems):
            self.write("## Problems per Schema\n")
            for name, problems in self.problems.items():
                self.write(f"### {name}")
                self.write_schema_problems(problems)

    def write_schema_problems(self, problems: list[LinterProblem]):
        errors = [p for p in problems if str(p.level) == RuleLevel.error.text]
        warnings = [p for p in problems if str(p.level) == RuleLevel.warning.text]
        if errors:
            self.write("#### Errors")
            for error in errors:
                self.write_problem(error)

        if warnings:
            self.write("#### Warnings")
            for warning in warnings:
                self.write_problem(warning)

        self.write("")

    def write_problem(self, problem: LinterProblem):
        formatted = "* "
        if problem.rule_name:
            formatted += problem.rule_name + ": "
        formatted += problem.message
        self.write(formatted)

    def end_report(self):
        self.write_summary()
        self.write_details()



import os
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

import click
import pydantic
from linkml_runtime.linkml_model.meta import ClassDefinitionName, SlotDefinition
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

MERMAID_SERIALIZATION = str


class Attribute(pydantic.BaseModel):
    datatype: str = "string"
    name: str = None
    key: str = None
    comment: str = None

    def __str__(self):
        cmt = f'"{self.comment}"' if self.comment else ""
        return f"    {self.datatype} {self.name} {self.key if self.key else ''} {cmt}"


class IdentifyingType(str, Enum):
    """Identifying types.

    See: https://mermaid.js.org/syntax/entityRelationshipDiagram.html#identification"""

    IDENTIFYING = "--"
    NON_IDENTIFYING = ".."

    def __str__(self):
        return self.value


class Cardinality(pydantic.BaseModel):
    """Cardinality of a slot.

    See https://mermaid.js.org/syntax/entityRelationshipDiagram.html#relationship-syntax"""

    required: bool = True
    multivalued: bool = False
    is_left: bool = True

    def __str__(self):
        required_char = "|" if self.required else "o"
        if self.multivalued:
            multivalued_char = "}" if self.is_left else "{"
        else:
            multivalued_char = "|"
        if self.is_left:
            return f"{multivalued_char}{required_char}"
        else:
            return f"{required_char}{multivalued_char}"


class RelationshipType(pydantic.BaseModel):
    """Relationship type.
    See https://mermaid.js.org/syntax/entityRelationshipDiagram.html#relationship-syntax"""

    left_cardinality: Cardinality
    identifying: IdentifyingType = IdentifyingType.IDENTIFYING
    right_cardinality: Cardinality = None

    def __str__(self):
        return f"{self.left_cardinality}{self.identifying}{self.right_cardinality}"


class Entity(pydantic.BaseModel):
    """Entity in an ER diagram."""

    name: str
    attributes: list[Attribute] = []

    def __str__(self):
        attrs = "\n".join([str(a) for a in self.attributes])
        return f"{self.name} {{\n{attrs}\n}}"


class Relationship(pydantic.BaseModel):
    """Relationship in an ER diagram."""

    first_entity: str
    relationship_type: RelationshipType = None
    second_entity: str = None
    relationship_label: str = None

    def __str__(self):
        return f'{self.first_entity} {self.relationship_type} {self.second_entity} : "{self.relationship_label}"'


class ERDiagram(pydantic.BaseModel):
    """Represents an Diagram of Entities and Relationships"""

    entities: list[Entity] = []
    relationships: list[Relationship] = []

    def __str__(self):
        # Sort entities and relationships for stable output
        sorted_entities = sorted(self.entities, key=lambda e: e.name)
        sorted_relationships = sorted(self.relationships, key=lambda r: str(r))

        ents = "\n".join([str(e) for e in sorted_entities])
        rels = "\n".join([str(r) for r in sorted_relationships])
        return f"erDiagram\n{ents}\n\n{rels}\n"


@dataclass
class ERDiagramGenerator(Generator):
    """
    A generator for serializing schemas as Entity-Relationship diagrams.

    Currently this generates diagrams in mermaid syntax, but in future
    this could easily be extended to have for example a direct SVG or PNG
    generator using PyGraphViz, similar to erdantic.
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.1"
    valid_formats = ["markdown", "mermaid"]
    uses_schemaloader = False
    requires_metamodel = False

    structural: bool = True
    """If True, then only the tree_root and entities reachable from the root are drawn"""

    exclude_attributes: bool = False
    """If True, do not include attributes in entities"""

    exclude_abstract_classes: bool = False
    """If True, do not include abstract classes in the diagram"""

    genmeta: bool = False
    gen_classvars: bool = True
    gen_slots: bool = True
    no_types_dir: bool = False
    use_slot_uris: bool = False

    preserve_names: bool = False
    """If true, preserve LinkML element names (classes/slots) in diagram labels."""

    def __post_init__(self):
        self.schemaview = SchemaView(self.schema)
        super().__post_init__()

    def serialize(self) -> MERMAID_SERIALIZATION:
        """
        Serialize a schema as an ER Diagram.

        If a tree_root is present in the schema, then only Entities traversable
        from here will be included. Otherwise, all Entities will be included.

        :return: mermaid string
        """
        sv = self.schemaview
        structural_roots = [cn for cn, c in sv.all_classes().items() if c.tree_root]
        if self.structural and structural_roots:
            return self.serialize_classes(structural_roots, follow_references=True)
        else:
            diagram = ERDiagram()
            for cn in sv.all_classes():
                self.add_class(cn, diagram)
            return self.serialize_diagram(diagram)

    def serialize_classes(
        self,
        class_names: list[Union[str, ClassDefinitionName]],
        follow_references=False,
        max_hops: int = None,
        include_upstream: bool = False,
    ) -> MERMAID_SERIALIZATION:
        """
        Serialize a list of classes as an ER Diagram.

        This will also traverse the reference graph and include any Entities reachable from the
        specified classes.

        By default, all reachable Entities are included, unless max_hops is specified.

        :param class_names: initial seed
        :param follow_references: if True, follow references even if not inlined
        :param max_hops: maximum number of hops to follow references
        :return:
        """
        visited = set()
        sv = self.schemaview
        stack = [(cn, 0) for cn in class_names]
        diagram = ERDiagram()
        while stack:
            cn, depth = stack.pop()
            if cn in visited:
                continue
            self.add_class(cn, diagram)
            visited.add(cn)
            if max_hops is not None and depth >= max_hops:
                continue
            for slot in sv.class_induced_slots(cn):
                rng = slot.range
                if rng in sv.all_classes():
                    if follow_references or sv.is_inlined(slot):
                        if rng not in visited:
                            stack.append((rng, depth + 1))

        # Now Add upstream classes if needed
        if include_upstream:
            for sn in sv.all_slots():
                slot = sv.schema.slots.get(sn)
                if slot and slot.range in set(class_names):
                    for cl in sv.all_classes():
                        if slot.name in sv.get_class(cl).slots and cl not in visited:
                            self.add_upstream_class(cl, set(class_names), diagram)
        return self.serialize_diagram(diagram)

    def serialize_diagram(self, diagram: ERDiagram) -> str:
        """
        Serialize an ER Diagram, from the native pydantic schema.

        :param diagram: ER Diagram
        :return:
        """
        er = str(diagram)
        if self.format == "markdown":
            return f"```mermaid\n{er}\n```\n"
        else:
            return er

    def add_upstream_class(self, class_name: ClassDefinitionName, targets: set[str], diagram: ERDiagram) -> None:
        sv = self.schemaview
        cls = sv.get_class(class_name)
        if self.exclude_abstract_classes and cls.abstract:
            return
        entity = Entity(name=cls.name if self.preserve_names else camelcase(cls.name))
        diagram.entities.append(entity)
        for slot in sv.class_induced_slots(class_name):
            if slot.range in targets:
                self.add_relationship(entity, slot, diagram)

    def add_class(self, class_name: ClassDefinitionName, diagram: ERDiagram) -> None:
        """
        Add a class to the ER Diagram.

        :param class_name: ClassDefinitionName
        :param diagram: ER Diagram
        :return:
        """
        sv = self.schemaview
        cls = sv.get_class(class_name)
        if self.exclude_abstract_classes and cls.abstract:
            return
        entity = Entity(name=cls.name if self.preserve_names else camelcase(cls.name))
        diagram.entities.append(entity)
        for slot in sv.class_induced_slots(class_name):
            # TODO: schemaview should infer this
            if slot.range is None:
                slot.range = sv.schema.default_range or "string"
            if slot.range in sv.all_classes():
                self.add_relationship(entity, slot, diagram)
            else:
                self.add_attribute(entity, slot)

    def add_relationship(self, entity: Entity, slot: SlotDefinition, diagram: ERDiagram) -> None:
        """
        Add a relationship to the ER Diagram.

        :param class_name: ClassDefinitionName
        :param slot: SlotDefinition
        :param diagram: ER Diagram
        :return:
        """
        sv = self.schemaview
        rel_type = RelationshipType(
            right_cardinality=Cardinality(
                required=slot.required is True, multivalued=slot.multivalued is True, is_left=True
            ),
            left_cardinality=Cardinality(is_left=False),
        )
        rel = Relationship(
            first_entity=entity.name,
            relationship_type=rel_type,
            second_entity=(
                sv.get_class(slot.range).name if self.preserve_names else camelcase(sv.get_class(slot.range).name)
            ),
            relationship_label=slot.name,
        )
        diagram.relationships.append(rel)

    def add_attribute(self, entity: Entity, slot: SlotDefinition) -> None:
        """
        Add an attribute to the ER Diagram.

        :param class_name: Class
        :param slot: SlotDefinition
        :return:
        """
        if self.exclude_attributes:
            return
        dt = slot.range
        if slot.multivalued:
            # NOTE: mermaid does not support []s or *s in attribute types
            dt = f"{dt}List"
        attr = Attribute(name=(slot.name if self.preserve_names else underscore(slot.name)), datatype=dt)
        entity.attributes.append(attr)


@shared_arguments(ERDiagramGenerator)
@click.option(
    "--structural/--no-structural",
    default=True,
    help="If True, then only the tree_root and entities reachable from the root are drawn",
)
@click.option(
    "--exclude-attributes/--no-exclude-attributes",
    default=False,
    help="If True, do not include attributes in entities",
)
@click.option(
    "--exclude-abstract-classes/--no-exclude-abstract-classes",
    default=False,
    help="If True, do not include abstract classes in the diagram",
)
@click.option(
    "--follow-references/--no-follow-references",
    default=False,
    help="If True, follow references even if not inlined",
)
@click.option("--max-hops", default=None, type=click.INT, help="Maximum number of hops")
@click.option("--classes", "-c", multiple=True, help="List of classes to serialize")
@click.option("--include-upstream", is_flag=True, help="Include upstream classes")
@click.version_option(__version__, "-V", "--version")
@click.command(name="erdiagram")
def cli(
    yamlfile,
    classes: list[str],
    max_hops: Optional[int],
    follow_references: bool,
    include_upstream: bool = False,
    **args,
):
    """Generate a mermaid ER diagram from a schema.

    By default, all entities traversable from the tree_root are included. If no tree_root is
    present, then all entities are included.

    To create an ER diagram for selected classes, use the --classes option.
    """
    gen = ERDiagramGenerator(
        yamlfile,
        **args,
    )
    if classes:
        print(
            gen.serialize_classes(
                classes,
                follow_references=follow_references,
                max_hops=max_hops,
                include_upstream=include_upstream,
            )
        )
    else:
        print(gen.serialize())


if __name__ == "__main__":
    cli()



import os
import re
from dataclasses import dataclass
from typing import Any, Callable, Optional, Union

import click
from jsonasobj2 import JsonObj, values
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ClassDefinitionName,
    Element,
    EnumDefinition,
    SlotDefinition,
    SubsetDefinition,
    TypeDefinition,
)
from linkml_runtime.utils.formatutils import be, camelcase, underscore

from linkml._version import __version__
from linkml.generators.yumlgen import YumlGenerator
from linkml.utils.deprecation import deprecation_warning
from linkml.utils.generator import Generator, shared_arguments
from linkml.utils.typereferences import References


@dataclass
class MarkdownGenerator(Generator):
    """
    Generates markdown documentation for a LinkML schema

    Each schema element (class, slot, type, enum) is translated into its own markdown file;
    additionally, an index.md is generated that links everything together.

    The markdown is suitable for deployment as a MkDocs or Sphinx site

    .. admonition:: Deprecated
        :class: warning

            The MarkdownGenerator class is being deprecated in favor of DocGenerator which can
            be found at the following path – `linkml/generators/docgen.py`. Going forward, DocGenerator
            which can be invoked using the `gen-doc` command will be the preferred way to generate
            Markdown documentation files for LinkML schemas.

            .. deprecated:: v1.9.1

            Recommendation: Update to use `gen-doc`
    """

    def __post_init__(self) -> None:
        deprecation_warning("gen-markdown")
        super().__post_init__()

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.2.1"
    directory_output = True
    valid_formats = ["md"]
    visit_all_class_slots = False
    uses_schemaloader = True

    # ObjectVars
    directory: Optional[str] = None
    image_directory: Optional[str] = None
    classes: set[ClassDefinitionName] = None
    image_dir: bool = False
    index_file: str = "index.md"
    noimages: bool = False
    noyuml: bool = False
    no_types_dir: bool = False
    warn_on_exist: bool = False
    gen_classes: Optional[set[ClassDefinitionName]] = None
    gen_classes_neighborhood: Optional[References] = None

    def visit_schema(
        self,
        directory: str = None,
        classes: set[ClassDefinitionName] = None,
        image_dir: bool = False,
        index_file: str = "index.md",
        noimages: bool = False,
        **_,
    ) -> str:
        self.gen_classes = classes if classes else []
        for cls in self.gen_classes:
            if cls not in self.schema.classes:
                raise ValueError("Unknown class name: {cls}")
        if self.gen_classes:
            self.gen_classes_neighborhood = self.neighborhood(list(self.gen_classes))

        self.directory = directory
        if directory:
            os.makedirs(directory, exist_ok=True)
        elif image_dir:
            raise ValueError("Image directory can only be used with '-d' option")
        if image_dir:
            self.image_directory = os.path.join(directory, "images")
            if not noimages:
                os.makedirs(self.image_directory, exist_ok=True)
        self.noimages = noimages
        if not self.no_types_dir:
            os.makedirs(os.path.join(directory, "types"), exist_ok=True)

        with open(self.exist_warning(directory, index_file), "w", encoding="UTF-8") as ixfile:
            items = []
            items.append(self.frontmatter(f"{self.schema.name}"))
            items.append(
                self.para(
                    f"**metamodel version:** {self.schema.metamodel_version}\n\n**version:** {self.schema.version}"
                )
            )
            items.append(self.para(be(self.schema.description)))

            items.append(self.header(3, "Classes"))
            for cls in sorted(self.schema.classes.values(), key=lambda c: c.name):
                if not cls.is_a and not cls.mixin and self.is_secondary_ref(cls.name):
                    items.append(self.class_hier(cls))

            items.append(self.header(3, "Mixins"))
            for cls in sorted(self.schema.classes.values(), key=lambda c: c.name):
                if cls.mixin and self.is_secondary_ref(cls.name):
                    items.append(self.class_hier(cls))

            items.append(self.header(3, "Slots"))
            for slot in sorted(self.schema.slots.values(), key=lambda s: s.name):
                if not slot.is_a and self.is_secondary_ref(slot.name):
                    items.append(self.pred_hier(slot))

            items.append(self.header(3, "Enums"))
            for enu in sorted(self.schema.enums.values(), key=lambda e: e.name):
                items.append(self.enum_hier(enu))

            items.append(self.header(3, "Subsets"))
            for subset in sorted(self.schema.subsets.values(), key=lambda s: s.name):
                items.append(self.bullet(self.subset_link(subset, use_desc=True), 0))

            items.append(self.header(3, "Types"))
            items.append(self.header(4, "Built in"))
            for builtin_name in sorted(self.synopsis.typebases.keys()):
                items.append(self.bullet(f"**{builtin_name}**"))
            items.append(self.header(4, "Defined"))
            for typ in sorted(self.schema.types.values(), key=lambda t: t.name):
                if self.is_secondary_ref(typ.name):
                    if typ.typeof:
                        typ_typ = self.type_link(typ.typeof)
                    else:
                        typ_typ = f"**{typ.base}**"

                    items.append(self.bullet(self.type_link(typ, after_link=f" ({typ_typ})", use_desc=True)))
            items = [i for i in items if i is not None]
            out = "\n".join(items) + "\n"
            out = pad_heading(out)
            out = out.rstrip() + "\n"
            ixfile.write(out)
        return out

    def visit_class(self, cls: ClassDefinition) -> str:
        # allow client to relabel metamodel
        mixin_local_name = self.get_metamodel_slot_name("Mixin")
        class_local_name = self.get_metamodel_slot_name("Class")

        if self.gen_classes and cls.name not in self.gen_classes:
            return ""

        with open(self.exist_warning(self.dir_path(cls)), "w", encoding="UTF-8") as clsfile:
            items = []
            class_curi = self.namespaces.uri_or_curie_for(str(self.namespaces._base), camelcase(cls.name))
            class_uri = self.namespaces.uri_for(class_curi)
            items.append(self.element_header(cls, cls.name, class_curi, class_uri))
            items.append("")
            if not self.noyuml:
                if self.image_directory:
                    yg = YumlGenerator(self)
                    yg.serialize(
                        classes=[cls.name],
                        directory=self.image_directory,
                        load_image=not self.noimages,
                    )
                    img_url = os.path.join("images", os.path.basename(yg.output_file_name))
                else:
                    yg = YumlGenerator(self)
                    img_url = (
                        yg.serialize(classes=[cls.name]).replace("?", "%3F").replace(" ", "%20").replace("|", "&#124;")
                    ).rstrip()

                items.append(f"[![img]({img_url})]({img_url})")

            if cls.id_prefixes:
                items.append(self.header(2, "Identifier prefixes"))
                for p in cls.id_prefixes:
                    items.append(self.bullet(f"{p}"))

            if cls.is_a is not None:
                items.append(self.header(2, "Parents"))
                items.append(self.bullet(f" is_a: {self.class_link(cls.is_a, use_desc=True)}"))
            if cls.mixins:
                items.append(self.header(2, f"Uses {mixin_local_name}"))
                for mixin in cls.mixins:
                    items.append(self.bullet(f" mixin: {self.class_link(mixin, use_desc=True)}"))

            if cls.name in self.synopsis.isarefs:
                items.append(self.header(2, "Children"))
                for child in sorted(self.synopsis.isarefs[cls.name].classrefs):
                    items.append(self.bullet(f"{self.class_link(child, use_desc=True)}"))

            if cls.name in self.synopsis.mixinrefs:
                items.append(self.header(2, f"{mixin_local_name} for"))
                for mixin in sorted(self.synopsis.mixinrefs[cls.name].classrefs):
                    items.append(self.bullet(f"{self.class_link(mixin, use_desc=True, after_link='(mixin)')}"))

            if cls.name in self.synopsis.classrefs:
                items.append(self.header(2, f"Referenced by {class_local_name}"))
                for sn in sorted(self.synopsis.classrefs[cls.name].slotrefs):
                    slot = self.schema.slots[sn]
                    if slot.range == cls.name:
                        items.append(
                            self.bullet(
                                f" **{self.class_link(slot.domain)}** "
                                f"*{self.slot_link(slot, add_subset=False)}*{self.predicate_cardinality(slot)}  "
                                f"**{self.class_type_link(slot.range)}**"
                            )
                        )

            items.append(self.header(2, "Attributes"))

            # List all of the slots that directly belong to the class
            slot_list = [slot for slot in [self.schema.slots[sn] for sn in cls.slots]]
            own_slots = [slot for slot in slot_list if cls.name in slot.domain_of]
            if own_slots:
                items.append(self.header(3, "Own"))
                for slot in own_slots:
                    items.append(self.slot_field(cls, slot))
                    slot_list.remove(slot)

            # List all of the inherited slots
            ancestors = set(self.ancestors(cls))
            inherited_slots = [slot for slot in slot_list if set(slot.domain_of).intersection(ancestors)]
            if inherited_slots:
                items.append(self.header(3, "Inherited from " + cls.is_a + ":"))
                for inherited_slot in inherited_slots:
                    items.append(self.slot_field(cls, inherited_slot))
                    slot_list.remove(inherited_slot)

            # List all of the slots acquired through mixing
            mixed_in_classes = set()
            for mixin in cls.mixins:
                mixed_in_classes.add(mixin)
                mixed_in_classes.update(set(self.ancestors(self.schema.classes[mixin])))
            for slot in slot_list:
                mixers = set(slot.domain_of).intersection(mixed_in_classes)
                for mixer in mixers:
                    items.append(self.header(3, "Mixed in from " + mixer + ":"))
                    items.append(self.slot_field(cls, slot))

            items.append(self.element_properties(cls))
            out = "\n".join(items)
            out = pad_heading(out)
            out = out.rstrip() + "\n"
            clsfile.write(out)
        return out

    def visit_type(self, typ: TypeDefinition) -> str:
        with open(self.exist_warning(self.dir_path(typ)), "w", encoding="UTF-8") as typefile:
            type_uri = typ.definition_uri
            type_curie = self.namespaces.curie_for(type_uri)
            out = self.element_header(typ, typ.name, type_curie, type_uri)

            out = "\n".join([out, "|  |  |  |"])
            out = "\n".join([out, "| --- | --- | --- |"])
            if typ.typeof:
                out = "\n".join([out, f"| Parent type | | {self.class_type_link(typ.typeof)} |"])
            out = "\n".join([out, f"| Root (builtin) type | | **{typ.base}** |"])
            if typ.repr:
                out = "\n".join([out, f"| Representation | | {typ.repr} |"])
            out += self.element_properties(typ)
            out += "\n"
            out = pad_heading(out)
            out = out.rstrip() + "\n"
            typefile.write(out)
        return out

    def visit_slot(self, aliased_slot_name: str, slot: SlotDefinition) -> str:
        with open(self.exist_warning(self.dir_path(slot)), "w", encoding="UTF-8") as slotfile:
            items = []
            slot_curie = self.namespaces.uri_or_curie_for(str(self.namespaces._base), underscore(slot.name))
            slot_uri = self.namespaces.uri_for(slot_curie)
            items.append(self.element_header(slot, aliased_slot_name, slot_curie, slot_uri))

            items.append(self.header(2, "Domain and Range"))
            items.append(
                f"{self.class_link(slot.domain)} &#8594;{self.predicate_cardinality(slot)} "
                f"{self.class_type_link(slot.range)}"
            )

            items.append(self.header(2, "Parents"))
            if slot.is_a:
                items.append(self.bullet(f" is_a: {self.slot_link(slot.is_a)}"))

            items.append(self.header(2, "Children"))
            if slot.name in sorted(self.synopsis.isarefs):
                for child in sorted(self.synopsis.isarefs[slot.name].slotrefs):
                    items.append(self.bullet(f" {self.slot_link(child)}"))

            items.append(self.header(2, "Used by"))
            if slot.name in sorted(self.synopsis.slotrefs):
                for rc in sorted(self.synopsis.slotrefs[slot.name].classrefs):
                    items.append(self.bullet(f"{self.class_link(rc)}"))
            if aliased_slot_name == "relation":
                if slot.subproperty_of:
                    reifies = (
                        self.slot_link(slot.subproperty_of)
                        if slot.subproperty_of in self.schema.slots
                        else slot.subproperty_of
                    )
                    items.append(self.bullet(f" reifies: {reifies}"))
            items.append(self.element_properties(slot))
            out = "\n".join(items)
            out = pad_heading(out)
            out = out.rstrip() + "\n"
            slotfile.write(out)
        return out

    def visit_enum(self, enum: EnumDefinition) -> str:
        with open(self.exist_warning(self.dir_path(enum)), "w", encoding="UTF-8") as enumfile:
            items = []
            enum_curie = self.namespaces.uri_or_curie_for(str(self.namespaces._base), underscore(enum.name))
            enum_uri = self.namespaces.uri_for(enum_curie)
            items.append(self.element_header(obj=enum, name=enum.name, curie=enum_curie, uri=enum_uri))
            items.append(self.element_properties(enum))
            out = "\n".join(items)
            out = pad_heading(out)
            out = out.rstrip() + "\n"
            enumfile.write(out)
        return out

    def visit_subset(self, subset: SubsetDefinition) -> str:
        with open(self.exist_warning(self.dir_path(subset)), "w", encoding="UTF-8") as subsetfile:
            items = []
            curie = self.namespaces.uri_or_curie_for(str(self.namespaces._base), underscore(subset.name))
            uri = self.namespaces.uri_for(curie)
            items.append(self.element_header(obj=subset, name=subset.name, curie=curie, uri=uri))
            # TODO: consider showing hierarchy within a subset
            items.append(self.header(3, "Classes"))
            for cls in sorted(self.schema.classes.values(), key=lambda c: c.name):
                if not cls.mixin:
                    if cls.in_subset and subset.name in cls.in_subset:
                        items.append(self.bullet(self.class_link(cls, use_desc=True), 0))
            items.append(self.header(3, "Mixins"))
            for cls in sorted(self.schema.classes.values(), key=lambda c: c.name):
                if cls.mixin:
                    if cls.in_subset and subset.name in cls.in_subset:
                        items.append(self.bullet(self.class_link(cls, use_desc=True), 0))
            items.append(self.header(3, "Slots"))
            for slot in sorted(self.schema.slots.values(), key=lambda s: s.name):
                if slot.in_subset and subset.name in slot.in_subset:
                    items.append(self.bullet(self.slot_link(slot, use_desc=True), 0))
            items.append(self.header(3, "Types"))
            for type in sorted(self.schema.types.values(), key=lambda s: s.name):
                if type.in_subset and subset.name in type.in_subset:
                    items.append(self.bullet(self.type_link(type, use_desc=True), 0))
            items.append(self.header(3, "Enums"))
            for enum in sorted(self.schema.enums.values(), key=lambda s: s.name):
                if enum.in_subset and subset.name in enum.in_subset:
                    items.append(self.bullet(self.enum_link(enum, use_desc=True), 0))
            items.append(self.element_properties(subset))
            out = "\n".join(items)
            out = pad_heading(out)
            out = out.rstrip() + "\n"
            subsetfile.write(out)
        return out

    def element_header(self, obj: Element, name: str, curie: str, uri: str) -> str:
        if isinstance(obj, TypeDefinition):
            obj_type = "Type"
        elif isinstance(obj, ClassDefinition):
            obj_type = "Class"
        elif isinstance(obj, SlotDefinition):
            obj_type = "Slot"
        elif isinstance(obj, EnumDefinition):
            obj_type = "Enum"
        elif isinstance(obj, SubsetDefinition):
            obj_type = "Subset"
        else:
            obj_type = "Class"

        header_label = f"{obj_type}: ~~{name}~~ _(deprecated)_" if obj.deprecated else f"{obj_type}: {name}"
        out = self.header(1, header_label)

        out += self.para(be(obj.description))
        out = "\n".join([out, f"URI: [{curie}]({uri})", ""])
        return out

    def element_properties(self, obj: Element) -> str:
        def identity(e: Any) -> Any:
            return e

        def prop_list(
            title: str,
            entries: Union[list, dict],
            formatter: Optional[Callable[[Element], str]] = None,
        ) -> Optional[str]:
            if formatter is None:
                formatter = identity
            if isinstance(entries, (dict, JsonObj)):
                entries = list(values(entries))
            if entries:
                items = []
                items.append(f"| **{title}:** | | {formatter(entries[0])} |")
                for entry in entries[1:]:
                    items.append(f"|  | | {formatter(entry)} |")
                return "\n".join(items)

        def enum_list(title: str, obj: EnumDefinition) -> str:
            # This data is from the enum provided in the YAML
            items = []
            items.append(self.header(2, title))
            items.append("| Text | Description | Meaning | Other Information |")
            items.append("| :--- | :---: | :---: | ---: |")

            for item, item_info in obj.permissible_values.items():
                text = ""
                desc = ""
                meaning = ""
                other = {}
                for k in item_info:
                    if item_info[k] is not None and len(item_info[k]) > 0:
                        if k == "text":
                            text = item_info[k]
                        elif k == "description":
                            desc = item_info[k]
                        elif k == "meaning":
                            meaning = item_info[k]
                        else:
                            other[k] = item_info[k]
                if not other:
                    other = ""
                items.append(f"| {text} | {desc} | {meaning} | {other} |")
            return "\n".join(items)

        items = []

        items.append(prop_list("Aliases", obj.aliases))
        items.append(
            prop_list(
                "Local names",
                obj.local_names,
                lambda e: f"{e.local_name_value} ({e.local_name_source})",
            )
        )
        items.append(prop_list("Mappings", obj.mappings))
        items.append(
            prop_list(
                "Alt Descriptions",
                obj.alt_descriptions,
                lambda e: f"{e.description} ({e.source})",
            )
        )
        # todos
        # notes
        items.append(prop_list("Comments", obj.comments))
        items.append(prop_list("Examples", obj.examples))
        items.append(prop_list("In Subsets", obj.in_subset))
        # from_schema
        # imported_from
        items.append(prop_list("See also", [f"[{v}]({v})" for v in obj.see_also]))
        items.append(prop_list("Exact Mappings", obj.exact_mappings))
        items.append(prop_list("Close Mappings", obj.close_mappings))
        items.append(prop_list("Narrow Mappings", obj.narrow_mappings))
        items.append(prop_list("Broad Mappings", obj.broad_mappings))
        items.append(prop_list("Related Mappings", obj.related_mappings))

        items = [i for i in items if i is not None]
        if len(items) > 0:
            header = "\n".join([self.header(2, "Other properties"), "|  |  |  |", "| --- | --- | --- |"])
            items.insert(0, header)

        #       - exact mappings
        #       - close mappings
        #       - related mappings
        #       - deprecated element has exact replacement
        #       - deprecated element has possible replacement
        if type(obj) is EnumDefinition:
            items.insert(0, enum_list("Permissible Values", obj))
            items.insert(1, "\n")

        out = "\n".join(items)
        return out

    def class_hier(self, cls: ClassDefinition, level=0) -> str:
        items = []
        items.append(self.bullet(self.class_link(cls, use_desc=True), level))
        if cls.name in sorted(self.synopsis.isarefs):
            for child in sorted(self.synopsis.isarefs[cls.name].classrefs):
                items.append(self.class_hier(self.schema.classes[child], level + 1))
        return "\n".join(items) if items else None

    def pred_hier(self, slot: SlotDefinition, level=0) -> str:
        items = []
        items.append(self.bullet(self.slot_link(slot, use_desc=True), level))
        if slot.name in sorted(self.synopsis.isarefs):
            for child in sorted(self.synopsis.isarefs[slot.name].slotrefs):
                items.append(self.pred_hier(self.schema.slots[child], level + 1))
        return "\n".join(items) if items else None

    def enum_hier(self, enum: EnumDefinition, level=0) -> str:
        items = []
        items.append(self.bullet(self.enum_link(enum, use_desc=True), level))
        if enum.name in sorted(self.synopsis.isarefs):
            for child in sorted(self.synopsis.isarefs[enum.name].classrefs):
                items.append(self.enum_hier(self.schema.enums[child], level + 1))
        return "\n".join(items) if items else None

    def dir_path(
        self,
        obj: Union[ClassDefinition, SlotDefinition, TypeDefinition, EnumDefinition],
    ) -> str:
        filename = (
            self.formatted_element_name(obj)
            if isinstance(obj, ClassDefinition)
            else (
                underscore(obj.name)
                if isinstance(obj, SlotDefinition)
                else underscore(obj.name)
                if isinstance(obj, EnumDefinition)
                else camelcase(obj.name)
            )
        )
        subdir = "/types" if isinstance(obj, TypeDefinition) and not self.no_types_dir else ""
        return f"{self.directory}{subdir}/{filename}.md"

    def is_secondary_ref(self, en: str) -> bool:
        """Determine whether 'en' is the name of something in the neighborhood of the requested classes

        @param en: element name
        @return: True if 'en' is the name of a slot, class or type in the immediate neighborhood of of what we are
        building
        """
        if not self.gen_classes:
            return True
        elif en in self.schema.classes:
            return en in self.gen_classes_neighborhood.classrefs
        elif en in self.schema.slots:
            return en in self.gen_classes_neighborhood.slotrefs
        elif en in self.schema.types:
            return en in self.gen_classes_neighborhood.typerefs
        else:
            return True

    def slot_field(self, cls: ClassDefinition, slot: SlotDefinition) -> str:
        items = []
        items.append(self.bullet(f"{self.slot_link(slot)}{self.predicate_cardinality(slot)}"))
        if slot.description:
            items.append(self.bullet(f"Description: {slot.description}", level=1))
        items.append(self.bullet(f"Range: {self.class_type_link(slot.range)}", level=1))
        # if slot.subproperty_of:
        #     self.bullet(f'edge label: {self.slot_link(slot.subproperty_of)}', level=1)
        for example in slot.examples:
            items.append(
                self.bullet(
                    f"Example: {getattr(example, 'value', ' ')} {getattr(example, 'description', ' ')}",
                    level=1,
                )
            )
        # if slot.name not in self.own_slot_names(cls):
        #     self.bullet(f'inherited from: {self.class_link(slot.domain)}', level=1)
        if slot.in_subset:
            ssl = ",".join(slot.in_subset)
            items.append(self.bullet(f"in subsets: ({ssl})", level=1))
        return "\n".join(items)

    def to_uri(self, uri_or_curie: str) -> str:
        """Return the URI for the slot if known"""
        if ":/" in uri_or_curie:
            return uri_or_curie
        elif self.namespaces.prefix_for(uri_or_curie) in self.namespaces:
            return self.namespaces.uri_for(uri_or_curie)
        else:
            return "*(Unknown namespace)*"

    # --
    # FORMATTING
    # --
    @staticmethod
    def predicate_cardinality(slot: SlotDefinition) -> str:
        """Emit cardinality for a suffix on a predicate"""
        if slot.multivalued:
            card_str = "1..\\*" if slot.required else "0..\\*"
        else:
            card_str = "1..1" if slot.required else "0..1"
        return f"  <sub>{card_str}</sub>"

    @staticmethod
    def range_cardinality(slot: SlotDefinition) -> str:
        """Emits cardinality decorator at end of type"""
        if slot.multivalued:
            card_str = "1..\\*" if slot.required else "0..\\*"
        else:
            card_str = "1..1" if slot.required else "0..1"
        return f"  <sub><b>{card_str}</b></sub>"

    @staticmethod
    def anchor(id_: str) -> str:
        return f'<a name="{id_}">'

    @staticmethod
    def anchorend() -> str:
        return "</a>"

    def header(self, level: int, txt: str) -> str:
        txt = self.get_metamodel_slot_name(txt)
        out = f"\n{'#' * level} {txt}\n"
        return out

    @staticmethod
    def para(txt: str) -> str:
        return f"\n{txt}\n"

    @staticmethod
    def bullet(txt: str, level=0) -> str:
        return f"{'    ' * level} * {txt}"

    def frontmatter(self, thingtype: str, layout="default") -> str:
        return self.header(1, thingtype)
        # print(f'---\nlayout: {layout}\n---\n')

    def bbin(self, obj: Element) -> str:
        """Boldify built in types

        @param obj: object name or id
        @return:
        """
        return obj.name if isinstance(obj, Element) else f"**{obj}**" if obj in self.synopsis.typebases else obj

    def desc_for(self, obj: Element, doing_descs: bool) -> str:
        """Return a description for object if it is unique (different than its parent)

        @param obj: object to be described
        @param doing_descs: If false, always return an empty string
        @return: text or empty string
        """
        if obj.description and doing_descs:
            if isinstance(obj, SlotDefinition) and obj.is_a:
                parent = self.schema.slots[obj.is_a]
            elif isinstance(obj, ClassDefinition) and obj.is_a:
                parent = self.schema.classes[obj.is_a]
            else:
                parent = None
            return "" if parent and obj.description == parent.description else obj.description
        return ""

    def _link(
        self,
        obj: Optional[Element],
        *,
        after_link: str = None,
        use_desc: bool = False,
        add_subset: bool = True,
    ) -> str:
        """Create a link to ref if appropriate.

        @param ref: the name or value of a class, slot, type or the name of a built in type.
        @param after_link: Text to put between link and description
        @param use_desc: True means append a description after the link if available
        @param add_subset: True means add any subset information that is available
        @return:
        """
        nl = "\n"
        if obj is None or not self.is_secondary_ref(obj.name):
            return self.bbin(obj)
        if isinstance(obj, SlotDefinition):
            link_name = ((be(obj.domain) + "➞") if obj.alias else "") + self.aliased_slot_name(obj)
            link_ref = underscore(obj.name)
        elif isinstance(obj, TypeDefinition):
            link_name = camelcase(obj.name)
            link_ref = f"types/{link_name}" if not self.no_types_dir else f"{link_name}"
        elif isinstance(obj, ClassDefinition):
            link_name = camelcase(obj.name)
            link_ref = camelcase(link_name)
        elif isinstance(obj, SubsetDefinition):
            link_name = camelcase(obj.name)
            link_ref = camelcase(link_name)
        else:
            link_name = obj.name
            link_ref = link_name
        desc = self.desc_for(obj, use_desc)
        return (
            f"[{link_name}]"
            f"({link_ref}.{self.format})"
            + (f" {after_link} " if after_link else "")
            + (f" - {desc.split(nl)[0]}" if desc else "")
        ).rstrip()

    def type_link(
        self,
        ref: Optional[Union[str, TypeDefinition]],
        *,
        after_link: str = None,
        use_desc: bool = False,
        add_subset: bool = True,
    ) -> str:
        return self._link(
            self.schema.types[ref] if isinstance(ref, str) else ref,
            after_link=after_link,
            use_desc=use_desc,
            add_subset=add_subset,
        )

    def slot_link(
        self,
        ref: Optional[Union[str, SlotDefinition]],
        *,
        after_link: str = None,
        use_desc: bool = False,
        add_subset: bool = True,
    ) -> str:
        return self._link(
            self.schema.slots[ref] if isinstance(ref, str) else ref,
            after_link=after_link,
            use_desc=use_desc,
            add_subset=add_subset,
        )

    def class_link(
        self,
        ref: Optional[Union[str, ClassDefinition]],
        *,
        after_link: str = None,
        use_desc: bool = False,
        add_subset: bool = True,
    ) -> str:
        return self._link(
            self.schema.classes[ref] if isinstance(ref, str) else ref,
            after_link=after_link,
            use_desc=use_desc,
            add_subset=add_subset,
        )

    def class_type_link(
        self,
        ref: Optional[Union[str, ClassDefinition, TypeDefinition, EnumDefinition]],
        *,
        after_link: str = None,
        use_desc: bool = False,
        add_subset: bool = True,
    ) -> str:
        if isinstance(ref, ClassDefinition):
            return self.class_link(ref, after_link=after_link, use_desc=use_desc, add_subset=add_subset)
        elif isinstance(ref, TypeDefinition):
            return self.type_link(ref, after_link=after_link, use_desc=use_desc, add_subset=add_subset)
        elif isinstance(ref, EnumDefinition):
            return self.type_link(ref, after_link=after_link, use_desc=use_desc, add_subset=add_subset)
        elif ref in self.schema.classes:
            return self.class_link(ref, after_link=after_link, use_desc=use_desc, add_subset=add_subset)
        elif ref in self.schema.enums:
            return self.enum_link(ref, after_link=after_link, use_desc=use_desc, add_subset=add_subset)
        else:
            return self.type_link(ref, after_link=after_link, use_desc=use_desc, add_subset=add_subset)

    def enum_link(
        self,
        ref: Optional[Union[str, EnumDefinition]],
        *,
        after_link: str = None,
        use_desc: bool = False,
        add_subset: bool = True,
    ) -> str:
        return self._link(
            self.schema.enums[ref] if isinstance(ref, str) else ref,
            after_link=after_link,
            use_desc=use_desc,
            add_subset=add_subset,
        )

    def subset_link(
        self,
        ref: Optional[Union[str, SubsetDefinition]],
        *,
        after_link: str = None,
        use_desc: bool = False,
    ) -> str:
        return self._link(
            self.schema.subsets[ref] if isinstance(ref, str) else ref,
            after_link=after_link,
            use_desc=use_desc,
        )

    def xlink(self, id_: str) -> str:
        return f"[{id_}]({self.id_to_url(id_)})"

    def exist_warning(self, *fpath: str) -> str:
        """Create a file name out of fpath and check whether it already exists"""
        fname = os.path.join(*fpath)
        if self.warn_on_exist and os.path.exists(fname):
            self.logger.warning(f"File {fname} already exists")
        return fname

    @staticmethod
    def id_to_url(id_: str) -> str:
        uri = id_
        if ":" in id_:
            # TODO! use PC
            if id_.startswith("SIO:"):
                uri = id_.replace("SIO:", "http://semanticscience.org/resource/SIO_")
            elif id_.startswith("HGNC:"):
                uri = "https://monarchinitiative.org/gene/" + id_
            else:
                frag = id_.replace(":", "_")
                base = "http://purl.obolibrary.org/obo/"
                uri = base + frag
        return uri


def pad_heading(text: str) -> str:
    """Add an extra newline to a non-top-level header that doesn't have one preceding it"""
    return re.sub(r"(?<!\n)\n##", "\n\n##", text)


@shared_arguments(MarkdownGenerator)
@click.command(name="markdown")
@click.option("--dir", "-d", required=True, help="Output directory")
@click.option("--classes", "-c", multiple=True, help="Class(es) to emit")
@click.option("--map-fields", "-M", multiple=True, help="Map metamodel fields, e.g. slot=field")
@click.option("--img", "-i", is_flag=True, help="Download YUML images to 'image' directory")
@click.option("--index-file", "-I", help="Name of markdown file that holds index")
@click.option("--noimages", is_flag=True, help="Do not (re-)generate images")
@click.option("--noyuml", is_flag=True, help="Do not add yUML figures to pages")
@click.option("--notypesdir", is_flag=True, help="Do not create a separate types directory")
@click.option("--warnonexist", is_flag=True, help="Warn if output file already exists")
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, map_fields, dir, img, index_file, notypesdir, warnonexist, **kwargs):
    """Generate markdown documentation of a LinkML model.

    .. warning::
        `gen-markdown` is deprecated. Please use `gen-doc` instead.
    """
    deprecation_warning("gen-markdown")

    gen = MarkdownGenerator(yamlfile, no_types_dir=notypesdir, warn_on_exist=warnonexist, **kwargs)
    if map_fields is not None:
        gen.metamodel_name_map = {}
        for mf in map_fields:
            [k, v] = mf.split("=")
            gen.metamodel_name_map[k] = v
    gen.serialize(directory=dir, image_dir=img, **kwargs)


if __name__ == "__main__":
    cli()



from abc import ABC, abstractmethod
from collections.abc import Iterator

from linkml.validator.report import ValidationResult
from linkml.validator.validation_context import ValidationContext


class ValidationPlugin(ABC):
    """Abstract base class for validation plugins.

    Subclasses must implement a ``process`` method.
    """

    def pre_process(self, context: ValidationContext) -> None:
        """A hook that will be called before instances are processed.

        :param context: A `ValidationContext` instance which provides
            access to the schema, target class, and artifacts generated
            from the schema
        """
        pass

    def post_process(self, context: ValidationContext) -> None:
        """A hook that will be called after instances are processed.

        :param context: A `ValidationContext` instance which provides
            access to the schema, target class, and artifacts generated
            from the schema
        """
        pass

    @abstractmethod
    def process(self, instance: dict, context: ValidationContext) -> Iterator[ValidationResult]:
        """Lazily yield validation results for an instance according to
        the validation context.

        :param instance: The instance to validate
        :param context: A `ValidationContext` instance which provides
            access to the schema, target class, and artifacts generated
            from the schema
        :return: Iterator over validation results
        :rtype: Iterator[ValidationResult]
        """
        pass



import os
import sys
from collections.abc import Iterable
from pathlib import Path

import click
import yaml

from linkml._version import __version__
from linkml.linter.config.datamodel.config import RuleLevel
from linkml.linter.formatters import JsonFormatter, MarkdownFormatter, TerminalFormatter, TsvFormatter
from linkml.linter.linter import Linter

YAML_SUFFIXES = [".yml", ".yaml"]
DEFAULT_CONFIG_FILES = [".linkmllint.yaml", ".linkmllint.yml"]


def get_yaml_files(root: Path, accept_dot_files: bool) -> Iterable[str]:
    if root.is_file():
        if root.suffix not in YAML_SUFFIXES:
            raise click.UsageError("SCHEMA must be a YAML file")
        yield str(root)
    else:
        for dir, _, files in os.walk(root):
            for file in files:
                if file in DEFAULT_CONFIG_FILES:
                    continue
                if file.startswith(".") and not accept_dot_files:
                    continue
                path = Path(dir, file)
                if path.suffix in YAML_SUFFIXES:
                    yield str(path)


@click.command(name="lint")
@click.argument(
    "schema",
    type=click.Path(exists=True, dir_okay=True, file_okay=True, resolve_path=True, path_type=Path),
)
@click.option("-a", "--all", is_flag=True, default=False, help="Process files that start with '.'.")
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Custom linter configuration file.",
)
@click.option(
    "-f",
    "--format",
    type=click.Choice(["terminal", "markdown", "json", "tsv"]),
    default="terminal",
    help="Report format.",
    show_default=True,
)
@click.option(
    "--validate",
    is_flag=True,
    default=False,
    help="Validate the schema against the LinkML Metamodel before linting.",
)
@click.option(
    "--validate-only",
    is_flag=True,
    default=False,
    help="Validate the schema against the LinkML Metamodel and then exit without checking linter rules.",
)
@click.option("-v", "--verbose", is_flag=True)
@click.option("-o", "--output", type=click.File("w"), default="-", help="Report file name.")
@click.option(
    "--ignore-warnings",
    is_flag=True,
    default=False,
    help="Do not exit with an error status if only warnings are found.",
)
@click.option(
    "--max-warnings",
    type=int,
    default=0,
    show_default=True,
    help="Do not exit with an error status if up to this number of warnings (and no errors) are found.",
)
@click.option("--fix/--no-fix", default=False)
@click.version_option(__version__, "-V", "--version")
def main(
    schema: Path,
    fix: bool,
    all: bool,
    config: str,
    format: str,
    validate: bool,
    validate_only: bool,
    output,
    ignore_warnings: bool,
    max_warnings: int,
    verbose: bool,
):
    """Run linter on SCHEMA.

    SCHEMA can be a single LinkML YAML file or a directory. If it is a directory
    every YAML file found in the directory (recursively) will be linted."""
    config_file = None
    if config:
        config_file = config
    else:
        for default_config in DEFAULT_CONFIG_FILES:
            path = Path(os.getcwd()) / default_config
            if path.exists():
                config_file = path
                break

    if config_file:
        with open(config_file) as f:
            config_dict = yaml.safe_load(f)
    else:
        config_dict = {"extends": "recommended"}

    linter = Linter(config_dict)
    if format == "terminal":
        formatter = TerminalFormatter(output, verbose)
    elif format == "markdown":
        formatter = MarkdownFormatter(output)
    elif format == "json":
        formatter = JsonFormatter(output)
    elif format == "tsv":
        formatter = TsvFormatter(output)

    error_count = 0
    warning_count = 0
    formatter.start_report()
    for path in get_yaml_files(schema, all):
        formatter.start_schema(path)
        report = linter.lint(path, fix=fix, validate_schema=validate, validate_only=validate_only)
        for problem in report:
            if str(problem.level) is RuleLevel.error.text:
                error_count += 1
            elif str(problem.level) is RuleLevel.warning.text:
                warning_count += 1
            formatter.handle_problem(problem)
        formatter.end_schema()
    formatter.end_report()

    exit_code = 0
    if error_count > 0:
        exit_code = 2
    elif not ignore_warnings and warning_count > max_warnings:
        exit_code = 1

    sys.exit(exit_code)


if __name__ == "__main__":
    main()



import json
import os
import warnings
from dataclasses import dataclass

import click
from linkml_runtime.linkml_model.meta import ClassDefinition, SlotDefinition
from linkml_runtime.utils.formatutils import be, camelcase, underscore

try:
    from terminusdb_client.woqlquery import WOQLQuery as WQ
except ImportError:
    WQ = None

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

# https://terminusdb.com/docs/terminusdb/#/reference/XSD_WHITELIST
XSD_Ok = {
    f"xsd:{t}"
    for t in [
        "string",
        "boolean",
        "decimal",
        "integer",
        "double",
        "float",
        "dateTime",
        "byte",
        "short",
        "integer",
        "long",
        "positiveInteger",
        "nonNegativeInteger",
        "negativeInteger",
        "nonPositiveInteger",
        "anyURI",
    ]
}


@dataclass
class TerminusdbGenerator(Generator):
    """
    Experimental generator for TerminusDB

    Generates JSON-LD to pass to `WOQLQuery()`.

    Assumes an "inference/main" graph if any slots have "is_a" values, because any statements with
    rdfs:subPropertyOf as the predicate must live in a TerminusDB "inference" graph rather than the
    "schema" graph. When creating a new TerminusDB database, only the "schema" and "instance" graphs
    are created. Thus, you may need to e.g. `WOQLClient.create_graph("inference", "main")`.

    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.0"
    valid_formats = ["json"]
    visit_all_class_slots = True
    uses_schemaloader = True

    # ObjectVars
    classes: list = None
    raw_additions: list = None
    clswq: str = None

    def __post_init__(self):
        if WQ is None:
            warnings.warn("terminusdb_client is not a requirement of this package, please install it separately")

    def visit_schema(self, inline: bool = False, **kwargs) -> None:
        self.classes = []
        self.raw_additions = []

    def end_schema(self, **_) -> str:
        return json.dumps(WQ().woql_and(*self.classes, *self.raw_additions).to_dict(), indent=2)

    def visit_class(self, cls: ClassDefinition) -> bool:
        self.clswq = WQ().add_class(camelcase(cls.name)).label(camelcase(cls.name)).description(be(cls.description))
        if cls.is_a:
            self.clswq.parent(camelcase(cls.is_a))
        if cls.abstract:
            self.clswq.abstract()
        if cls.broad_mappings:
            if any(
                str(self.namespaces.uri_for(m)) == "http://terminusdb.com/schema/system#Document"
                for m in cls.broad_mappings
            ):
                self.clswq.parent("Document")
        return True

    def end_class(self, cls: ClassDefinition) -> None:
        self.classes.append(self.clswq)

    def visit_class_slot(self, cls: ClassDefinition, aliased_slot_name: str, slot: SlotDefinition) -> None:
        if slot.range in self.schema.classes:
            rng = camelcase(slot.range)
        elif slot.range in self.schema.types:
            # XXX Why does `linkml_runtime.utils.metamodelcore.Identifier` subclass `str`??
            rng = str(self.schema.types[slot.range].uri)
        else:
            rng = "xsd:string"

        name = f"{cls.name} {aliased_slot_name}" if slot.is_usage_slot else aliased_slot_name

        # translate to terminusdb xsd builtins:
        if rng == "xsd:int":
            rng = "xsd:integer"
        elif rng == "xsd:float":
            rng = "xsd:double"
        elif rng == "xsd:language":
            rng = "xsd:string"

        if rng not in XSD_Ok and slot.range not in self.schema.classes:
            raise Exception(
                f"slot range for {name} must be schema class or supported xsd type. Range {rng} is of type {type(rng)}."
            )

        self.clswq.property(underscore(name), rng, label=name, description=slot.description)
        if not slot.multivalued:
            self.clswq.max(1)
        if slot.required:
            self.clswq.min(1)
        if slot.is_a:
            self.raw_additions.append(
                WQ().add_quad(
                    underscore(name),
                    "rdfs:subPropertyOf",
                    self.clswq.iri(underscore(slot.is_a)),
                    "inference/main",
                )
            )


@shared_arguments(TerminusdbGenerator)
@click.version_option(__version__, "-V", "--version")
@click.command(name="terminusdb")
def cli(yamlfile, **args):
    """Generate graphql representation of a LinkML model"""
    print(TerminusdbGenerator(yamlfile, **args).serialize(**args))


if __name__ == "__main__":
    cli()



sqlalchemy_imperative_template_str = """
from dataclasses import dataclass
from dataclasses import field
from typing import List

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import Text
from sqlalchemy import Integer
from sqlalchemy.orm import registry
from sqlalchemy.orm import relationship

mapper_registry = registry()
metadata = MetaData()

{% if not no_model_import %}
from {{model_path}} import *
{% endif %}

{% for c in classes %}
tbl_{{classname(c.name)}} = Table('{{c.name}}', metadata,
    {%- for s in c.attributes.values() %}
    Column('{{s.name}}',
          Text,
          {% if 'foreign_key' in s.annotations -%}
            ForeignKey('{{ s.annotations['foreign_key'].value }}'),
          {% endif -%}
          {% if 'primary_key' in s.annotations -%}
            primary_key=True
          {%- endif -%}
          ),
    {%- endfor %}
)
{% endfor -%}

# -- Mappings --

{% for c in classes if not is_join_table(c) %}
mapper_registry.map_imperatively({{classname(c.name)}}, tbl_{{classname(c.name)}}, properties = {
  ## NOTE: mapping omitted for now, see https://stackoverflow.com/questions/11746922/sqlalchemy-object-has-no-attribute-sa-adapter
  {% for mapping in backrefs[c.name] %}
  {% if mapping.uses_join_table %}
  {% else %}
  #'{{mapping.source_slot}}': relationship( {{ mapping.target_class }}, backref='{{c.name}}' ),
  {% endif %}
  #'{mapping.source_slot}': relationship()
  ## {{ mapping }}
  {% endfor %}
})
{% endfor %}
"""



"""
Generate GOlr YAML schema definitions.

These can be converted to solr schema-xml, and used in amigo-bbop tools

See the golr-views directory in this repo for examples

"""

import os
from dataclasses import dataclass
from typing import Optional

import click
from linkml_runtime.linkml_model.meta import ClassDefinition, SlotDefinition
from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.utils.metamodelcore import empty_list
from linkml_runtime.utils.yamlutils import YAMLRoot, as_yaml

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class GOLRField(YAMLRoot):
    id: str
    description: str
    display_name: str
    property: list = empty_list()
    cardinality: Optional[str] = None


@dataclass
class GOLRClass(YAMLRoot):
    id: str
    schema_generating: bool
    description: str
    display_name: str
    document_category: str
    weight: int
    fields: list[GOLRField] = empty_list()


@dataclass
class GolrSchemaGenerator(Generator):
    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    directory_output = True
    valid_formats = ["golr"]
    visit_all_class_slots = True
    uses_schemaloader = True
    requires_metamodel = False

    # ObjectVars
    directory: str = None
    class_obj: Optional[GOLRClass] = None

    def generate_header(self):
        headers = [f"# metamodel_version: {self.schema.metamodel_version}"]
        if self.schema.version:
            headers.append(f"# version: {self.schema.version}")
        return headers

    def visit_schema(self, directory: str, **_) -> None:
        self.directory = directory
        if directory:
            os.makedirs(directory, exist_ok=True)
        # write_golr_yaml_to_dir(schema, dir)

    def visit_class(self, cls: ClassDefinition) -> bool:
        if not (cls.mixin or cls.abstract):
            self.class_obj = GOLRClass(
                id=underscore(cls.name),
                schema_generating=True,
                description=cls.description,
                display_name=cls.name,
                document_category=cls.name,
                weight=20,
            )
            return True
        else:
            return False

    def end_class(self, cls: ClassDefinition) -> None:
        fn = os.path.join(self.directory, underscore(cls.name + "-config.yaml"))
        if len(self.class_obj.fields) > 1:
            with open(fn, "w", encoding="UTF-8") as f:
                f.write("".join(self.generate_header()))
                f.write(as_yaml(self.class_obj))

    def visit_class_slot(self, cls: ClassDefinition, aliased_slot_name: str, slot: SlotDefinition) -> None:
        field = GOLRField(
            id=underscore(aliased_slot_name),
            description=slot.description,
            display_name=slot.name,
        )
        if slot.multivalued:
            field.cardinality = "multi"
        self.class_obj.fields.append(field)


@shared_arguments(GolrSchemaGenerator)
@click.command(name="golr-views")
@click.option("--dir", "-d", default="golr-views", show_default=True, help="Output directory")
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, dir=None, **args):
    """Generate GOLR representation of a LinkML model"""
    print(GolrSchemaGenerator(yamlfile, directory=dir, **args).serialize(directory=dir, **args))


if __name__ == "__main__":
    cli()



from linkml.validators.jsonschemavalidator import JsonSchemaDataValidator
from linkml.validators.sparqlvalidator import SparqlDataValidator

__all__ = ["JsonSchemaDataValidator", "SparqlDataValidator"]



"""
Custom exceptions
"""


class SchemaError(ValueError):
    """Base class for errors relating to schema specification, parsing, structure, etc."""


class ValidationError(SchemaError):
    """Schema is invalid!"""



import dataclasses
import logging
import os
from copy import deepcopy
from pathlib import Path
from typing import Optional, Union, cast
from urllib.parse import urlparse

from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    Element,
    EnumDefinition,
    SchemaDefinition,
    SlotDefinition,
    SlotDefinitionName,
    TypeDefinition,
    TypeDefinitionName,
)
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.namespaces import Namespaces
from linkml_runtime.utils.yamlutils import extended_str
from rdflib import URIRef

logger = logging.getLogger(__name__)


def merge_schemas(
    target: SchemaDefinition,
    mergee: SchemaDefinition,
    imported_from: Optional[str] = None,
    namespaces: Optional[Namespaces] = None,
    merge_imports: bool = True,
) -> None:
    """Merge mergee into target"""
    assert target.name is not None, "Schema name must be supplied"
    if target.license is None:
        target.license = mergee.license

    resolve_merged_imports(target, mergee, imported_from)
    set_from_schema(mergee)

    if namespaces:
        merge_namespaces(target, mergee, namespaces)

    if merge_imports:
        for prefix in mergee.emit_prefixes:
            if prefix not in target.emit_prefixes:
                target.emit_prefixes.append(prefix)

    if imported_from is None:
        imported_from_uri = None
    else:
        if imported_from.startswith("http") or ":" not in imported_from:
            imported_from_uri = imported_from
        else:
            imported_from_uri = namespaces.uri_for(imported_from)
    merge_dicts(target.classes, mergee.classes, imported_from, imported_from_uri, merge_imports)
    merge_dicts(target.slots, mergee.slots, imported_from, imported_from_uri, merge_imports)
    merge_dicts(target.types, mergee.types, imported_from, imported_from_uri, merge_imports)
    merge_dicts(target.subsets, mergee.subsets, imported_from, imported_from_uri, merge_imports)
    merge_dicts(target.enums, mergee.enums, imported_from, imported_from_uri, merge_imports)


def merge_namespaces(target: SchemaDefinition, mergee: SchemaDefinition, namespaces) -> None:
    """
    Add the mergee namespace definitions to target

    :param target:
    :param mergee:
    :param namespaces:
    :return:
    """
    for prefix in mergee.prefixes.values():
        # Handle local prefixes special: we assume that these happen because we are in different (levels of) folders,
        # and we assume that they reference the same linkml file
        if "://" not in prefix.prefix_reference:
            # We cannot resolve this to an absolute path, so we have to assume that
            # this prefix is already defined correctly in the target
            if prefix.prefix_prefix not in namespaces:
                logger.info(
                    "Adding an unadjusted relative prefix for %s from %s, "
                    + "as the prefix is not yet defined, even as we cannot adjust it relative to the final file. "
                    + "If it cannot be resolved, add the prefix definition to the input schema!",
                    prefix.prefix_prefix,
                    mergee.name,
                )
                namespaces[prefix.prefix_prefix] = prefix.prefix_reference
            else:
                if (
                    prefix.prefix_prefix in target.prefixes
                    and target.prefixes[prefix.prefix_prefix].prefix_reference != prefix.prefix_reference
                ):
                    logger.info(
                        "Ignoring different relative prefix for %s from %s, "
                        + "as we cannot adjust it relative to the final file. "
                        + "Assuming the first found location is correct: %s!",
                        prefix.prefix_prefix,
                        mergee.name,
                        namespaces[prefix.prefix_prefix],
                    )
            continue

        namespaces[prefix.prefix_prefix] = prefix.prefix_reference
        # if prefix.prefix_prefix not in target.prefixes:
        #     target.prefixes[prefix.prefix_prefix] = prefix
        if (
            prefix.prefix_prefix in target.prefixes
            and target.prefixes[prefix.prefix_prefix].prefix_reference != prefix.prefix_reference
        ):
            raise ValueError(f"Prefix: {prefix.prefix_prefix} mismatch between {target.name} and {mergee.name}")
    for mmap in mergee.default_curi_maps:
        namespaces.add_prefixmap(mmap)


def set_from_schema(schema: SchemaDefinition) -> None:
    for t in [schema.subsets, schema.classes, schema.slots, schema.types, schema.enums]:
        for k in t.keys():
            t[k].from_schema = schema.id
            if isinstance(t[k], SlotDefinition):
                fragment = underscore(t[k].name)
            else:
                fragment = camelcase(t[k].name)
            if schema.default_prefix in schema.prefixes:
                ns = schema.prefixes[schema.default_prefix].prefix_reference
            else:
                ns = str(URIRef(schema.id) + "/")
            t[k].definition_uri = f"{ns}{fragment}"


def merge_dicts(
    target: dict[str, Element],
    source: dict[str, Element],
    imported_from: str,
    imported_from_uri: str,
    merge_imports: bool,
) -> None:
    for k, v in source.items():
        if k in target and source[k].from_schema != target[k].from_schema:
            raise ValueError(f"Conflicting URIs ({source[k].from_schema}, {target[k].from_schema}) for item: {k}")
        target[k] = deepcopy(v)
        # currently all imports closures are merged into main schema, EXCEPT
        # internal linkml types, which are considered separate
        # https://github.com/linkml/issues/121
        if imported_from is not None:
            if (
                not merge_imports
                or imported_from.startswith("linkml")
                or imported_from_uri.startswith("https://w3id.org/biolink/linkml")
            ):
                target[k].imported_from = imported_from


def merge_slots(
    target: Union[SlotDefinition, TypeDefinition],
    source: Union[SlotDefinition, TypeDefinition],
    skip: list[Union[SlotDefinitionName, TypeDefinitionName]] = None,
    inheriting: bool = True,
) -> None:
    """
    Merge slot source into target

    :param target: slot to merge into
    :param source: slot to be merged from
    :param skip: Properties to not merge (used to prevent provenance such as 'inherited from' from propagating)
    :param inheriting: True means source is the parent.  False means that everything gets copied
    """
    if skip is None:
        skip = []
    for k, v in dataclasses.asdict(source).items():
        if k not in skip and v is not None and (not inheriting or getattr(target, k, None) is None):
            if k in source._inherited_slots or not inheriting:
                setattr(target, k, deepcopy(v))
            else:
                setattr(target, k, None)
    target.__post_init__()


def slot_usage_name(usage_name: SlotDefinitionName, owning_class: ClassDefinition) -> SlotDefinitionName:
    """
     Synthesize a unique name for an overridden slot

    :param usage_name:
    :param owning_class:
    :return: Synthesized name
    """
    return SlotDefinitionName(extended_str.concat(owning_class.name, "_", usage_name))


def alias_root(schema: SchemaDefinition, slotname: SlotDefinitionName) -> Optional[SlotDefinitionName]:
    """Return the ultimate alias of a slot"""
    alias = schema.slots[slotname].alias if slotname in schema.slots else None
    if alias and alias == slotname:
        raise ValueError("Error: Slot {slotname} is aliased to itself.")
    return alias_root(schema, cast(SlotDefinitionName, alias)) if alias else slotname


def merge_classes(
    schema: SchemaDefinition,
    target: ClassDefinition,
    source: ClassDefinition,
    at_end: bool = False,
) -> None:
    """Merge the slots in source into target

    :param schema: Containing schema
    :param target: mergee
    :param source: class to merge
    :param at_end: True means add mergee to the end.  False to the front
    """

    # List of grounded slots referenced in the target class
    target_base_slots = set(alias_root(schema, s) for s in target.slots)

    for slotname in source.slots if at_end else source.slots[::-1]:
        slotbase = alias_root(schema, slotname)
        if slotbase in target.slot_usage:
            slotname = slot_usage_name(slotbase, target)
        if slotbase not in target_base_slots:
            target.slots.append(slotname) if at_end else target.slots.insert(0, slotname)
            target_base_slots.add(slotbase)


def merge_enums(
    schema: SchemaDefinition,
    target: EnumDefinition,
    source: EnumDefinition,
    at_end: bool = False,
) -> None:
    """Merge the slots in source into target

    :param schema: Containing schema
    :param target: mergee
    :param source: enum to merge
    :param at_end: True means add mergee to the end.  False to the front
    """
    # TODO: Finish enumeration merge code
    pass


def resolve_merged_imports(
    target: SchemaDefinition, mergee: SchemaDefinition, imported_from: Optional[str] = None
) -> None:
    """Merge the imports in source into target

    :param target: Containing schema
    :param mergee: mergee
    :param imported_from: file that the mergee was imported from
    :param at_end: True means add mergee to the end.  False to the front
    """

    for imp in mergee.imports:
        if imp is None:
            continue

        # Skip if already imported
        if imp in target.imports:
            continue

        # Leave as-is if the import has a scheme (e.g. http://)
        elif urlparse(imp).scheme:
            target.imports.append(imp)
            continue

        # Leave as-is if the import is an absolute path
        elif Path(imp).is_absolute():
            target.imports.append(imp)
            continue

        # Adjust relative imports to be relative to the importing file
        elif imp.startswith("."):
            if imported_from is None:
                logger.warning(f"Cannot resolve relative import: {imp}")
                target.imports.append(imp)
            else:
                resolved_imp = os.path.normpath(str(Path(imported_from).parent / Path(imp)))
                target.imports.append(resolved_imp)

        # By default, add the import as-is
        else:
            target.imports.append(imp)



from __future__ import annotations

import os
from typing import Any, TextIO

import click
import yaml
from linkml_runtime.linkml_model.meta import ClassDefinition, SchemaDefinition
from linkml_runtime.utils.schemaview import SchemaView

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

DEFAULT_SOURCE_JSON = "data.json~jsonpath"
DEFAULT_ITERATOR = "$.items[*]"


class YarrrmlGenerator(Generator):
    generatorname = os.path.basename(__file__)
    generatorversion = "0.2.1"
    valid_formats = ["yml", "yaml"]
    visit_all_class_slots = False

    def __init__(
        self,
        schema: str | TextIO | SchemaDefinition,
        format: str = "yml",
        **kwargs,
    ):
        raw_src = kwargs.pop("source", None)
        it = kwargs.pop("iterator_template", None)
        super().__init__(schema, **kwargs)

        self.schemaview = SchemaView(schema)
        self.schema: SchemaDefinition = self.schemaview.schema
        self.format = format
        self.source: str = self._infer_source_suffix(raw_src) if raw_src else DEFAULT_SOURCE_JSON
        self.iterator_template: str = it or DEFAULT_ITERATOR

    def _infer_source_suffix(self, path: str) -> str:
        p = (path or "").lower()
        if "~" in p:
            return path
        if p.endswith(".json"):
            return f"{path}~jsonpath"
        if p.endswith(".csv") or p.endswith(".tsv"):
            return f"{path}~csv"
        return path

    def serialize(self, **args) -> str:
        data = yaml.safe_dump(
            self.as_dict(), sort_keys=False, allow_unicode=True, default_flow_style=False, indent=2, width=120
        )
        return data

    def as_dict(self) -> dict[str, Any]:
        sv = self.schemaview
        mappings: dict[str, Any] = {}

        for cls in sv.all_classes().values():
            mapping_dict = {}

            if self._is_json_source():
                mapping_dict["sources"] = [[self.source, self._iterator_for_class(cls)]]
            else:
                mapping_dict["sources"] = [[self.source]]

            mapping_dict["s"] = self._subject_template_for_class(cls)

            mapping_dict["po"] = self._po_list_for_class(cls)

            mappings[str(cls.name)] = mapping_dict

        result = {"prefixes": self._prefixes(), "mappings": mappings}
        return result

    # helpers
    def _is_json_source(self) -> bool:
        return "~jsonpath" in (self.source or "")

    def _prefixes(self) -> dict[str, str]:
        px: dict[str, str] = {}
        if self.schema.prefixes:
            for p in self.schema.prefixes.values():
                if p.prefix_prefix and p.prefix_reference:
                    px[str(p.prefix_prefix)] = str(p.prefix_reference)
        px.setdefault("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#")
        return px

    def _iterator_for_class(self, c: ClassDefinition) -> str:
        # supports {Class}
        return self.iterator_template.replace("{Class}", c.name)

    def _subject_template_for_class(self, c: ClassDefinition) -> str:
        sv = self.schemaview
        default_prefix = sv.schema.default_prefix or "ex"
        id_slot = sv.get_identifier_slot(c.name)
        if id_slot:
            return f"{default_prefix}:$({id_slot.name})"
        key_slot = sv.get_key_slot(c.name)
        if key_slot:
            return f"{default_prefix}:$({key_slot.name})"
        return f"{default_prefix}:{c.name}/$(subject_id)"

    def _po_list_for_class(self, c: ClassDefinition) -> list[dict[str, Any]]:
        sv = self.schemaview
        po: list[dict[str, Any]] = []

        class_curie = sv.get_uri(c, expand=False)
        if class_curie is not None:
            po.append({"p": "rdf:type", "o": str(class_curie)})

        default_prefix = sv.schema.default_prefix or "ex"

        for s in sv.class_induced_slots(c.name):
            pred_uri = sv.get_uri(s, expand=False)
            pred = str(pred_uri) if pred_uri is not None else f"{default_prefix}:{s.name}"

            decl = sv.get_slot(s.name)
            alias = decl.alias if decl and decl.alias else s.alias
            var = alias or s.name

            is_obj = sv.get_class(s.range) is not None if s.range else False
            if is_obj:
                inlined = None
                if decl and decl.inlined is not None:
                    inlined = decl.inlined
                if inlined is False:
                    po.append({"p": pred, "o": {"value": f"$({var})", "type": "iri"}})
                continue

            po.append({"p": pred, "o": f"$({var})"})
        return po


@shared_arguments(YarrrmlGenerator)
@click.command(name="yarrrml")
@click.option(
    "--source",
    help="YARRRML source shorthand, e.g., data.json~jsonpath or data.csv~csv (TSV works too)",
)
@click.option(
    "--iterator-template",
    help='JSONPath iterator template; supports {Class}, default: "$.items[*]"',
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, source, iterator_template, **args):
    """Generate YARRRML mappings from a LinkML schema."""
    if source:
        args["source"] = source
    if iterator_template:
        args["iterator_template"] = iterator_template
    gen = YarrrmlGenerator(yamlfile, **args)
    print(gen.serialize(**args))


if __name__ == "__main__":
    cli()



"""
Main ``linkml`` entrypoint

Gathers all the other linkml click entrypoints and puts them under ``linkml`` :)
"""

import click

from linkml._version import __version__
from linkml.converter.cli import cli as linkml_convert
from linkml.generators.csvgen import cli as gen_csv
from linkml.generators.dbmlgen import cli as gen_dbml
from linkml.generators.docgen import cli as gen_doc
from linkml.generators.dotgen import cli as gen_graphviz
from linkml.generators.erdiagramgen import cli as gen_erdiagram
from linkml.generators.excelgen import cli as gen_excel
from linkml.generators.golanggen import cli as gen_golang
from linkml.generators.golrgen import cli as gen_golr_views
from linkml.generators.graphqlgen import cli as gen_graphql
from linkml.generators.javagen import cli as gen_java
from linkml.generators.jsonldcontextgen import cli as gen_jsonld_context
from linkml.generators.jsonldgen import cli as gen_jsonld
from linkml.generators.jsonschemagen import cli as gen_json_schema
from linkml.generators.linkmlgen import cli as gen_linkml
from linkml.generators.markdowngen import cli as gen_markdown
from linkml.generators.namespacegen import cli as gen_namespaces
from linkml.generators.owlgen import cli as gen_owl
from linkml.generators.panderagen import cli as gen_pandera
from linkml.generators.plantumlgen import cli as gen_plantuml
from linkml.generators.prefixmapgen import cli as gen_prefix_map
from linkml.generators.projectgen import cli as gen_project
from linkml.generators.protogen import cli as gen_proto
from linkml.generators.pydanticgen import cli as gen_pydantic
from linkml.generators.pythongen import cli as gen_python
from linkml.generators.rdfgen import cli as gen_rdf
from linkml.generators.rustgen.cli import cli as gen_rust
from linkml.generators.shaclgen import cli as gen_shacl
from linkml.generators.shexgen import cli as gen_shex
from linkml.generators.sparqlgen import cli as gen_sparql
from linkml.generators.sqlalchemygen import cli as gen_sqla
from linkml.generators.sqltablegen import cli as gen_sqltables
from linkml.generators.sssomgen import cli as gen_sssom
from linkml.generators.summarygen import cli as gen_summary
from linkml.generators.terminusdbgen import cli as gen_terminusdb
from linkml.generators.typescriptgen import cli as gen_typescript
from linkml.generators.yamlgen import cli as gen_yaml
from linkml.generators.yarrrmlgen import cli as gen_yarrrml
from linkml.generators.yumlgen import cli as gen_yuml
from linkml.linter.cli import main as linkml_lint
from linkml.utils.execute_tutorial import cli as run_tutorial
from linkml.utils.schema_fixer import main as linkml_schema_fixer
from linkml.utils.sqlutils import main as linkml_sqldb
from linkml.validator.cli import cli as linkml_validate
from linkml.workspaces.example_runner import cli as linkml_run_examples

# --------------------------------------------------
# Command groups
# --------------------------------------------------


@click.group()
@click.version_option(__version__, "-V", "--version")
def linkml():
    """
    LinkML: A flexible linked data modeling language
    """


@linkml.group()
@click.version_option(__version__, "-V", "--version")
def generate():
    """
    Generate formats from a LinkML schema
    """


@linkml.group()
@click.version_option(__version__, "-V", "--version")
def dev():
    """
    Helper tools for linkml development
    """


# --------------------------------------------------
# Add commands to groups
# --------------------------------------------------

# Top-level linkml commands
linkml.add_command(linkml_convert, name="convert")
linkml.add_command(linkml_lint, name="lint")
linkml.add_command(linkml_sqldb, name="sqldb")
linkml.add_command(linkml_schema_fixer, name="fix")
linkml.add_command(linkml_run_examples, name="examples")
linkml.add_command(linkml_validate, name="validate")

# Generators
generate.add_command(gen_jsonld_context, name="jsonld-context")
generate.add_command(gen_prefix_map, name="prefix-map")
generate.add_command(gen_csv, name="csv")
generate.add_command(gen_graphviz, name="graphviz")
generate.add_command(gen_golang, name="golang")
generate.add_command(gen_golr_views, name="golr-views")
generate.add_command(gen_graphql, name="graphql")
generate.add_command(gen_java, name="java")
generate.add_command(gen_jsonld, name="jsonld")
generate.add_command(gen_json_schema, name="json-schema")
generate.add_command(gen_markdown, name="markdown")
generate.add_command(gen_doc, name="doc")
generate.add_command(gen_namespaces, name="namespaces")
generate.add_command(gen_owl, name="owl")
generate.add_command(gen_plantuml, name="plantuml")
generate.add_command(gen_proto, name="proto")
generate.add_command(gen_python, name="python")
generate.add_command(gen_pydantic, name="pydantic")
generate.add_command(gen_pandera, name="pandera")
generate.add_command(gen_rdf, name="rdf")
generate.add_command(gen_rust, name="rust")
generate.add_command(gen_shex, name="shex")
generate.add_command(gen_shacl, name="shacl")
generate.add_command(gen_sparql, name="sparql")
generate.add_command(gen_typescript, name="typescript")
generate.add_command(gen_terminusdb, name="terminusdb")
generate.add_command(gen_yuml, name="yuml")
generate.add_command(gen_yaml, name="yaml")
generate.add_command(gen_erdiagram, name="erdiagram")
generate.add_command(gen_sqla, name="sqla")
generate.add_command(gen_sqltables, name="sqltables")
generate.add_command(gen_summary, name="summary")
generate.add_command(gen_project, name="project")
generate.add_command(gen_excel, name="excel")
generate.add_command(gen_sssom, name="sssom")
generate.add_command(gen_linkml, name="linkml")
generate.add_command(gen_dbml, name="dbml")
generate.add_command(gen_yarrrml, name="yarrrml")

# Dev helpers
dev.add_command(run_tutorial, name="tutorial")
